# -*- coding: utf-8 -*-
class DataPairs:
    """
    Bundles :attr:`observations` with :attr:`ground_truth`.

    Attributes
    ----------
    observations : list of observation space elements
        The observations, possibly distorted or low-dimensional.
    ground_truth: list of reconstruction space elements or `None`
        The ground truth data (may be replaced with good quality references).
        If not known, it may be omitted (`None`).
    """
    def __init__(self, observations, ground_truth=None, name='',
                 description=''):
        self.observations = observations
        if not isinstance(self.observations, list):
            self.observations = [self.observations]
        self.ground_truth = ground_truth
        if (self.ground_truth is not None and
                not isinstance(self.ground_truth, list)):
            self.ground_truth = [self.ground_truth]
        self.name = name
        self.description = description

    def __repr__(self):
        return ("DataPairs(observations=\n{observations}, "
                "ground_truth=\n{ground_truth}, name='{name}', "
                "description='{description}')".format(
                    observations=self.observations,
                    ground_truth=self.ground_truth,
                    name=self.name,
                    description=self.description))

    def __str__(self):
        return ("DataPairs '{name}'".format(name=self.name) if self.name
                else self.__repr__())

    def __len__(self):
        return len(self.observations)



# -*- coding: utf-8 -*-
"""Provides the abstract reconstructor base class."""
from inspect import signature, Parameter


class Reconstructor:
    """Abstract reconstructor base class.

    There are two ways of implementing a `Reconstructor` subclass:

        * Implement :meth:`reconstruct`. It has to support optional in-place
          and out-of-place evaluation.
        * Implement :meth:`_reconstruct`. It must have one of the following
          signatures:

            - ``_reconstruct(self, observation, out)`` (in-place)
            - ``_reconstruct(self, observation)`` (out-of-place)
            - ``_reconstruct(self, observation, out=None)`` (optional
              in-place)

    Attributes
    ----------
    reco_space : :class:`odl.discr.DiscretizedSpace`, optional
        Reconstruction space.
    observation_space : :class:`odl.discr.DiscretizedSpace`, optional
        Observation space.
    name : str
        Name of the reconstructor.
    hyper_params : dict
        Current hyper parameter values.
        Initialized automatically using the default values from
        :attr:`HYPER_PARAMS` (but may be overridden by `hyper_params` passed
        to :meth:`__init__`).
    """

    HYPER_PARAMS = {}
    """Specification of hyper parameters.

    This class attribute is a dict that lists the hyper parameter of the
    reconstructor.
    It should not be hidden by an instance attribute of the same name (i.e. by
    assigning a value to `self.HYPER_PARAMS` in an instance of a subtype).

    The keys of this dict are the names of the hyper parameters, and each
    value is a dict with the following fields.

        Standard fields:

            ``'default'``
                Default value.
            ``'retrain'`` : bool, optional
                Whether training depends on the parameter. Default: ``False``.
                Any custom subclass of `LearnedReconstructor` must set this
                field to ``True`` if training depends on the parameter value.

        Hyper parameter search fields:

            ``'range'`` : (float, float), optional
                Interval of valid values. If this field is set, the parameter
                is taken to be real-valued.
                Either ``'range'`` or ``'choices'`` has to be set.
            ``'choices'`` : sequence, optional
                Sequence of valid values of any type. If this field is set,
                ``'range'`` is ignored. Can be used to perform manual grid
                search. Either ``'range'`` or ``'choices'`` has to be set.
            ``'method'`` : {'grid_search', 'hyperopt'}, optional
                 Optimization method for the parameter.
                 Default: ``'grid_search'``.
                 Options are:

                    ``'grid_search'``
                        Grid search over a sequence of fixed values. Can be
                        configured by the dict ``'grid_search_options'``.
                    ``'hyperopt'``
                        Random search using the ``hyperopt`` package. Can be
                        configured by the dict ``'hyperopt_options'``.

            ``'grid_search_options'`` : dict
                Option dict for grid search with the fields:

                    ``'num_samples'`` : int, optional
                        Number of values. Default: 10.
                        Only relevant if ``'range'`` is specified and
                        ``'choices'`` is not.
                    ``'type'`` : {'linear'}, optional
                        Type of grid, i.e. distribution of the values.
                        Default: ``'linear'``.
                        Only relevant if ``'range'`` is specified and
                        ``'choices'`` is not.
                        Options are:

                            ``'linear'``
                                Equidistant values in the ``'range'``.

            ``'hyperopt_options'`` : dict
                Option dict for ``'hyperopt'`` method with the fields:

                    ``'space'`` : hyperopt space, optional
                        Custom hyperopt search space. If this field is set,
                        ``'range'`` and ``'type'`` are ignored.
                    ``'type'`` : {'uniform'}, optional
                        Type of the space for sampling. Default: ``'uniform'``.
                        Options are:

                            ``'uniform'``
                                Uniform distribution over the ``'range'``.
    """

    def __init__(self, reco_space=None, observation_space=None,
                 name='', hyper_params=None):
        self.reco_space = reco_space
        self.observation_space = observation_space
        self.name = name or self.__class__.__name__
        self.hyper_params = {k: v['default']
                             for k, v in self.HYPER_PARAMS.items()}
        if hyper_params is not None:
            self.hyper_params.update(hyper_params)

    def reconstruct(self, observation, out=None):
        """Reconstruct input data from observation data.

        The default implementation calls `_reconstruct`, automatically choosing
        in-place or out-of-place evaluation.

        Parameters
        ----------
        observation : :attr:`observation_space` element-like
            The observation data.
        out : :attr:`reco_space` element-like, optional
            Array to which the result is written (in-place evaluation).
            If `None`, a new array is created (out-of-place evaluation).

        Returns
        -------
        reconstruction : :attr:`reco_space` element or `out`
            The reconstruction.
        """
        parameters = signature(self._reconstruct).parameters
        if 'out' in parameters:
            if out is not None:
                self._reconstruct(observation, out)
                reco = out
            elif parameters['out'].default == Parameter.empty:
                reco = self.reco_space.zero()
                self._reconstruct(observation, reco)
            else:
                reco = self._reconstruct(observation)
        else:
            reco = self._reconstruct(observation)
            if out is not None:
                out[:] = reco
        return reco

    def _reconstruct(self, observation, *args, **kwargs):
        """Reconstruct input data from observation data.

        This method must have one of the following signatures:
            - ``_reconstruct(self, observation, out)`` (in-place)
            - ``_reconstruct(self, observation)`` (out-of-place)
            - ``_reconstruct(self, observation, out=None)`` (optional
              in-place)

        The parameters and return value are documented in :meth:`reconstruct`.
        """
        raise NotImplementedError("'_reconstruct' not implemented by class "
                                  "'{}'. Reconstructor subclasses must "
                                  "implement either 'reconstruct' or "
                                  "'_reconstruct'.".format(type(self)))


class LearnedReconstructor(Reconstructor):
    def train(self, dataset):
        """Train the reconstructor with a dataset by adapting its parameters.

        Should only use the training data from `dataset`.

        Parameters
        ----------
        dataset : :class:`.Dataset`
            The dataset from which the training data should be used.
        """
        raise NotImplementedError

    def save_params(self, path):
        """Save parameters to path.

        Parameters
        ----------
        path : str
            Path at which the parameters should be saved.
            Implementations should either use it as a file path (w/ or w/o
            extension) or as a directory path for multiple files (the dir
            should be created by this method if it does not exist).
        """
        raise NotImplementedError

    def load_params(self, path):
        """Load parameters from path.

        Parameters
        ----------
        path : str
            Path at which the parameters are stored. C.f. `save_params`.
        """
        raise NotImplementedError


class IterativeReconstructor(Reconstructor):
    """Iterative reconstructor base class.

    Subclasses should call ``self.callback()`` after each iteration in
    ``self.reconstruct``.

    Attributes
    ----------
    callback : ``odl.solvers.util.callback.Callback`` or `None`
        Callback to be called after each iteration.
    """
    def __init__(self, callback=None, **kwargs):
        self.callback = None
        super().__init__(**kwargs)


class FunctionReconstructor(Reconstructor):
    """Reconstructor defined by a callable.

    Attributes
    ----------
    function : callable
        Callable that is used in `reconstruct`.
    fun_args : list
        Arguments to be passed to `function`.
    fun_kwargs : dict
        Keyword arguments to be passed to `function`.
    """
    def __init__(self, function, *args, fun_args=None, fun_kwargs=None,
                 **kwargs):
        """
        Parameters
        ----------
        function : callable
            Callable that is used in :meth:`reconstruct`.
        fun_args : list, optional
            Arguments to be passed to `function`.
        fun_kwargs : dict, optional
            Keyword arguments to be passed to `function`.
        """
        super().__init__(*args, **kwargs)
        self.function = function
        self.fun_args = fun_args or []
        self.fun_kwargs = fun_kwargs or {}

    def _reconstruct(self, observation):
        return self.function(observation, *self.fun_args, **self.fun_kwargs)



# -*- coding: utf-8 -*-
"""
Optimization of hyper parameters.

Both grid search and random search using the ``hyperopt`` library are
supported.

The hyper parameter specification of a reconstructor class, optionally
including default options for optimization, are specified in the class
attribute :attr:`~dival.Reconstructor.HYPER_PARAMS`.
"""
from itertools import product
from warnings import warn
import numpy as np
from hyperopt import hp, fmin, tpe, Trials, space_eval
from tqdm import tqdm
from dival.util.std_out_err_redirect_tqdm import std_out_err_redirect_tqdm
from dival.measure import Measure
from dival import LearnedReconstructor


def optimize_hyper_params(reconstructor, validation_data, measure,
                          dataset=None,
                          HYPER_PARAMS_override=None,
                          hyperopt_max_evals=1000,
                          hyperopt_max_evals_retrain=1000,
                          hyperopt_rstate=None,
                          show_progressbar=True,
                          tqdm_file=None):
    """Optimize hyper parameters of a reconstructor.

    Parameters
    ----------
    reconstructor : :class:`.Reconstructor`
        The reconstructor.
    validation_data : :class:`.DataPairs`
        The test data on which the performance is measured.
    measure : :class:`.Measure` or str
        The measure to use as the objective. The sign is chosen automatically
        depending on the measures :attr:`~Measure.measure_type`.
    dataset : :class:`.Dataset`, optional
        The dataset used for training `reconstructor` if it is a
        :class:`LearnedReconstructor`.
    HYPER_PARAMS_override : dict, optional
        Hyper parameter specification overriding the defaults
        in ``type(reconstructor).HYPER_PARAMS``.
        The structure of this dict is the same as the structure
        of :attr:`Reconstructor.HYPER_PARAMS`, except that all
        fields are optional.
        Here, each value of a dict for one parameter is treated
        as an entity, i.e. specifying the dict
        ``HYPER_PARAMS[...]['grid_search_options']`` overrides
        the whole dict, not only the specified keys in it.
    hyperopt_max_evals : int, optional
        Number of evaluations for different combinations of the parameters that
        are optimized by ``hyperopt`` and that do not require retraining.
        Should be chosen depending on the complexity of dependence and the
        number of such parameters.
    hyperopt_max_evals_retrain : int, optional
        Number of evaluations for different combinations of the parameters that
        are optimized by ``hyperopt`` and that require retraining.
        Should be chosen depending on the complexity of dependence and the
        number of such parameters.
    hyperopt_rstate : :class:`np.random.RandomState`, optional
        Random state for the random searches performed by ``hyperopt``.
    show_progressbar : bool, optional
        Whether to show a progress bar for the optimization. Default: ``True``.
    tqdm_file : file-like object
        File/stream to pass to ``tqdm``.
    """
    if isinstance(measure, str):
        measure = Measure.get_by_short_name(measure)
    if dataset is None and isinstance(reconstructor, LearnedReconstructor):
        raise ValueError('dataset required for training of '
                         '`LearnedReconstructor`')

    if HYPER_PARAMS_override is None:
        HYPER_PARAMS_override = {}
    for k in HYPER_PARAMS_override.keys():
        if k not in type(reconstructor).HYPER_PARAMS.keys():
            warn("unknown hyper param '{}' for reconstructor of type '{}'"
                 .format(k, type(reconstructor)))

    params = {}
    params_retrain = {}
    for k, v in type(reconstructor).HYPER_PARAMS.items():
        param = v.copy()
        param.update(HYPER_PARAMS_override.get(k, {}))
        param.setdefault('method', 'grid_search')
        retrain = v.get('retrain', False)
        if retrain:
            params_retrain[k] = param
        else:
            params[k] = param

    loss_sign = 1 if measure.measure_type == 'distance' else -1

    def fn(x):
        reconstructor.hyper_params.update(x)
        reconstructions = [reconstructor.reconstruct(observation) for
                           observation in validation_data.observations]
        measure_values = [measure.apply(r, g) for r, g in
                          zip(reconstructions, validation_data.ground_truth)]
        loss = loss_sign * np.mean(measure_values)

        return {'status': 'ok',
                'loss': loss}

    def fn_retrain(x):
        reconstructor.hyper_params.update(x)
        reconstructor.train(dataset)

        best_sub_hp = _optimize_hyper_params_impl(
            reconstructor, fn, params,
            hyperopt_max_evals=hyperopt_max_evals,
            hyperopt_rstate=hyperopt_rstate, show_progressbar=False)

        reconstructions = [reconstructor.reconstruct(observation) for
                           observation in validation_data.observations]
        measure_values = [measure.apply(r, g) for r, g in
                          zip(reconstructions, validation_data.ground_truth)]
        loss = loss_sign * np.mean(measure_values)

        return {'status': 'ok',
                'loss': loss,
                'best_sub_hp': best_sub_hp}

    if params_retrain:
        best_hyper_params = _optimize_hyper_params_impl(
            reconstructor, fn_retrain, params_retrain,
            hyperopt_max_evals=hyperopt_max_evals_retrain,
            hyperopt_rstate=hyperopt_rstate,
            show_progressbar=show_progressbar, tqdm_file=tqdm_file)
    else:
        best_hyper_params = _optimize_hyper_params_impl(
            reconstructor, fn, params,
            hyperopt_max_evals=hyperopt_max_evals,
            hyperopt_rstate=hyperopt_rstate,
            show_progressbar=show_progressbar, tqdm_file=tqdm_file)

    return best_hyper_params


def _optimize_hyper_params_impl(reconstructor, fn, params,
                                hyperopt_max_evals=1000, hyperopt_rstate=None,
                                show_progressbar=True, tqdm_file=None):
    grid_search_params = []
    grid_search_param_choices = []
    hyperopt_space = {}
    for k, param in params.items():
        method = param['method']
        if method == 'grid_search':
            grid_search_options = param.get('grid_search_options', {})
            choices = param.get('choices')
            if choices is None:
                range_ = param.get('range')
                if range_ is not None:
                    grid_type = grid_search_options.get('type', 'linear')
                    if grid_type == 'linear':
                        n = grid_search_options.get('num_samples', 10)
                        choices = np.linspace(range_[0], range_[1], n)
                    else:
                        raise ValueError(
                            "unknown grid type '{grid_type}' in {reco_cls}."
                            "HYPER_PARAMS['{k}']['grid_search_options']".
                            format(
                                grid_type=grid_type,
                                reco_cls=reconstructor.__class__.__name__,
                                k=k))
                else:
                    raise ValueError(
                        "neither 'choices' nor 'range' is specified in "
                        "{reco_cls}.HYPER_PARAMS['{k}'], one of them must be "
                        "specified for grid search".format(
                            reco_cls=reconstructor.__class__.__name__, k=k))
            grid_search_params.append(k)
            grid_search_param_choices.append(choices)
        elif method == 'hyperopt':
            hyperopt_options = param.get('hyperopt_options', {})
            space = hyperopt_options.get('space')
            if space is None:
                choices = param.get('choices')
                if choices is None:
                    range_ = param.get('range')
                    if range_ is not None:
                        space_type = hyperopt_options.get('type', 'uniform')
                        if space_type == 'uniform':
                            space = hp.uniform(k, range_[0], range_[1])
                        else:
                            raise ValueError(
                                "unknown hyperopt space type '{space_type}' "
                                "in {reco_cls}.HYPER_PARAMS['{k}']"
                                "['hyperopt_options']".format(
                                    space_type=space_type,
                                    reco_cls=reconstructor.__class__.__name__,
                                    k=k))
                    else:
                        raise ValueError(
                            "neither 'choices' nor 'range' is specified in "
                            "{reco_cls}.HYPER_PARAMS['{k}']"
                            "['hyperopt_options']. One of these or "
                            "{reco_cls}.HYPER_PARAMS['{k}']"
                            "['hyperopt_options']['space'] must be specified "
                            "for hyperopt param search".format(
                                reco_cls=reconstructor.__class__.__name__,
                                k=k))
                else:
                    space = hp.choice(k, choices)
            hyperopt_space[k] = space
        else:
            raise ValueError("unknown method '{method}' for "
                             "{reco_cls}.HYPER_PARAMS['{k}']".format(
                                 method=method,
                                 reco_cls=reconstructor.__class__.__name__,
                                 k=k))

    best_loss = np.inf

    best_hyper_params = None
    with std_out_err_redirect_tqdm(tqdm_file) as orig_stdout:
        grid_search_total = np.prod([len(c) for c in
                                     grid_search_param_choices])
        for grid_search_values in tqdm(
                product(*grid_search_param_choices),
                desc='hyper param opt. for {reco_cls}'.format(
                    reco_cls=type(reconstructor).__name__),
                total=grid_search_total,
                file=orig_stdout,
                leave=False,
                disable=not show_progressbar):
            grid_search_param_dict = dict(zip(grid_search_params,
                                              grid_search_values))
            reconstructor.hyper_params.update(grid_search_param_dict)
            if len(hyperopt_space) == 0:
                result = fn({})
                if result['loss'] < best_loss:
                    best_loss = result['loss']
                    best_hyper_params = result.get('best_sub_hp', {})
                    best_hyper_params.update(grid_search_param_dict)
            else:
                trials = Trials()
                argmin = fmin(fn=fn, space=hyperopt_space, algo=tpe.suggest,
                              max_evals=hyperopt_max_evals, trials=trials,
                              rstate=hyperopt_rstate,
                              show_progressbar=False)
                best_trial = trials.best_trial
                if best_trial['result']['loss'] < best_loss:
                    best_loss = best_trial['result']['loss']
                    best_hyper_params = best_trial['result'].get('best_sub_hp',
                                                                 {})
                    best_hyper_params.update(grid_search_param_dict)
                    best_hyper_params.update(space_eval(hyperopt_space,
                                                        argmin))

    if best_hyper_params is not None:
        reconstructor.hyper_params.update(best_hyper_params)

    return best_hyper_params



# -*- coding: utf-8 -*-
"""Implements datasets for training and evaluating learned reconstructors.

.. autosummary::
    get_standard_dataset
    Dataset
    GroundTruthDataset
    ObservationGroundTruthPairDataset
    EllipsesDataset
    LoDoPaBDataset

The function :func:`.get_standard_dataset` returns fixed "standard" datasets
with pairs of observation and ground truth samples.
Currently the standard datasets are ``'ellipses'`` and ``'lodopab'``.

The class :class:`.ObservationGroundTruthPairDataset` can be used, either
directly or via :meth:`.GroundTruthDataset.create_pair_dataset`, to create a
custom dataset of pairs given a ground truth dataset and a forward operator.
For example:

    * define a :class:`.GroundTruthDataset` object (e.g. \
      :class:`.EllipsesDataset`)
    * define a forward operator
    * call :meth:`~.GroundTruthDataset.create_pair_dataset` of the dataset and
      pass the forward operator as well as some noise specification if wanted
"""
from warnings import warn

__all__ = ['get_standard_dataset', 'Dataset', 'GroundTruthDataset',
           'ObservationGroundTruthPairDataset', 'EllipsesDataset',
           'LoDoPaBDataset']

from .standard import get_standard_dataset
from .dataset import (Dataset, GroundTruthDataset,
                      ObservationGroundTruthPairDataset)
from .ellipses_dataset import EllipsesDataset
try:
    from .lodopab_dataset import LoDoPaBDataset
except FileNotFoundError as e:
    __all__.remove('LoDoPaBDataset')
    warn('could not import `LoDoPaBDataset` because of the following '
         'error:\n\n{}\n'.format(e))



# -*- coding: utf-8 -*-
"""Provides standard datasets for benchmarking.
"""
import numpy as np
from skimage.transform import resize
import odl
from dival.datasets.ellipses_dataset import EllipsesDataset
from dival.datasets.lodopab_dataset import LoDoPaBDataset


STANDARD_DATASET_NAMES = ['ellipses', 'lodopab']


def get_standard_dataset(name, **kwargs):
    """
    Return a standard dataset by name.

    The standard datasets are (currently):

        ``'ellipses'``
            A typical synthetical CT dataset with ellipse phantoms.

            `EllipsesDataset` is used as ground truth dataset, a ray
            transform with parallel beam geometry using 30 angles is applied,
            and white gaussian noise with a standard deviation of 5% (i.e.
            ``0.05 * mean(abs(observation))``) is added. The ray transform is
            normalized by its spectral norm.

            A normalized ray transform that corresponds to the noiseless
            forward operator is stored in the attribute `ray_trafo` of the
            dataset.

            In order to avoid the inverse crime, the ground truth images of
            shape (128, 128) are upscaled by bilinear interpolation to a
            resolution of (400, 400) before the ray transform is applied (whose
            discretization is different from the one of `ray_trafo`).

        ``'lodopab'``
            A simulated low dose CT dataset based on real reconstructions from
            the `LIDC-IDRI
            <https://wiki.cancerimagingarchive.net/display/Public/LIDC-IDRI>`_
            dataset.

            The dataset contains 42895 pairs of images and projection data.
            For simulation, a ray transform with parallel beam geometry using
            1000 angles and 513 detector pixels is used. Poisson noise
            corresponding to 4096 incident photons per pixel before attenuation
            is applied to the projection data.

            A ray transform that corresponds to the noiseless forward operator
            is available via
            :meth:`~dival.datasets.LoDoPaBDataset.get_ray_trafo` from this
            dataset.

    Parameters
    ----------
    name : str
        Name of the dataset.
    kwargs : dict
        Keyword arguments.
        Supported parameters for the datasets are:

            ``'ellipses'``
                no parameters
            ``'lodopab'``
                observation_model : {``'post-log'``, ``'pre-log'``}, optional
                    The observation model to use. Default is ``'post-log'``.
                min_photon_count : float, optional
                    Replacement value for a simulated photon count of zero.
                    If ``observation_model == 'post-log'``, a value greater
                    than zero is required in order to avoid undefined values.
                    The default is 0.1, both for ``'post-log'`` and
                    ``'pre-log'`` model.

    Returns
    -------
    dataset : :class:`.Dataset`
        The standard dataset.
        It has an attribute `standard_dataset_name` that stores its name.
    """
    name = name.lower()
    if name == 'ellipses':
        ellipses_dataset = EllipsesDataset(image_size=128)

        NUM_ANGLES = 30
        # image shape for simulation
        IM_SHAPE = (400, 400)  # images will be scaled up from (128, 128)

        reco_space = ellipses_dataset.space
        space = odl.uniform_discr(min_pt=reco_space.min_pt,
                                  max_pt=reco_space.max_pt,
                                  shape=IM_SHAPE, dtype=np.float32)

        reco_geometry = odl.tomo.parallel_beam_geometry(
            reco_space, num_angles=NUM_ANGLES)
        geometry = odl.tomo.parallel_beam_geometry(
            space, num_angles=NUM_ANGLES,
            det_shape=reco_geometry.detector.shape)

        impl = kwargs.pop('impl', 'astra_cuda')
        ray_trafo = odl.tomo.RayTransform(space, geometry, impl=impl)

        class _ResizeOperator(odl.Operator):
            def __init__(self):
                super().__init__(reco_space, space)

            def _call(self, x, out):
                out.assign(space.element(resize(x, IM_SHAPE, order=1)))

        # Ensure operator has fixed operator norm for scale invariance
        opnorm = odl.power_method_opnorm(ray_trafo)
        resize_op = _ResizeOperator()
        forward_op = (1 / opnorm) * ray_trafo * resize_op

        dataset = ellipses_dataset.create_pair_dataset(
            forward_op=forward_op, noise_type='white',
            noise_kwargs={'relative_stddev': True,
                          'stddev': 0.05},
            noise_seeds={'train': 1, 'validation': 2, 'test': 3})

        def get_ray_trafo(impl=impl):
            return odl.tomo.RayTransform(reco_space, reco_geometry, impl=impl)

        dataset.get_ray_trafo = get_ray_trafo
        dataset.ray_trafo = get_ray_trafo(impl=impl)

    elif name == 'lodopab':
        dataset = LoDoPaBDataset(**kwargs)

    else:
        raise ValueError("unknown dataset '{}'. Known standard datasets are {}"
                         .format(name, STANDARD_DATASET_NAMES))

    dataset.standard_dataset_name = name
    return dataset



# -*- coding: utf-8 -*-
"""Provides `EllipsesDataset`."""
from itertools import repeat
import numpy as np
from odl.discr.lp_discr import uniform_discr
from odl.phantom import ellipsoid_phantom
from dival.datasets.dataset import GroundTruthDataset


class EllipsesDataset(GroundTruthDataset):
    """Dataset with images of multiple random ellipses.

    This dataset uses :meth:`odl.phantom.ellipsoid_phantom` to create
    the images.
    The images are normalized to have a value range of ``[0., 1.]``.

    Attributes
    ----------
    space
        ``odl.uniform_discr(min_pt, max_pt, (image_size, image_size),
        dtype='float32')``, with the parameters passed to :meth:`__init__`.
    shape
        ``(image_size, image_size)``, with `image_size` parameter passed to
        :meth:`__init__`. Default ``(128, 128)``.
    train_len
        `train_len` parameter passed to :meth:`__init__`.
        Default ``32000``.
    validation_len
        `validation_len` parameter passed to :meth:`__init__`.
        Default ``3200``.
    test_len
        `test_len` parameter passed to :meth:`__init__`.
        Default ``3200``.
    random_access
        ``False``
    num_elements_per_sample
        ``1``
    """
    def __init__(self, image_size=128, min_pt=None, max_pt=None,
                 train_len=32000, validation_len=3200, test_len=3200,
                 fixed_seeds=False):
        """
        Parameters
        ----------
        image_size : int, optional
            Number of pixels per image dimension. Default: ``128``.
        min_pt : [int, int], optional
            Minimum values of the lp space.
            Default: ``[-image_size/2, -image_size/2]``.
        max_pt : [int, int], optional
            Maximum values of the lp space.
            Default: ``[image_size/2, image_size/2]``.
        train_len : int or `None`, optional
            Length of training set. Default: ``32000``.
            If `None`, infinitely many samples could be generated.
        validation_len : int, optional
            Length of training set. Default: ``3200``.
        test_len : int, optional
            Length of test set. Default: ``3200``.
        fixed_seeds : dict or bool, optional
            Seeds to use for random generation.
            The values of the keys ``'train'``, ``'validation'`` and ``'test'``
            are used. If a seed is `None` or omitted, it is choosen randomly.
            If ``True`` is passed, the seeds
            ``fixed_seeds={'train': 42, 'validation': 2, 'test': 1}`` are used.
            If ``False`` is passed (the default), all seeds are chosen
            randomly.
        """
        self.shape = (image_size, image_size)
        if min_pt is None:
            min_pt = [-self.shape[0]/2, -self.shape[1]/2]
        if max_pt is None:
            max_pt = [self.shape[0]/2, self.shape[1]/2]
        space = uniform_discr(min_pt, max_pt, self.shape, dtype=np.float32)
        self.train_len = train_len
        self.validation_len = validation_len
        self.test_len = test_len
        self.random_access = False
        if isinstance(fixed_seeds, bool):
            if fixed_seeds:
                self.fixed_seeds = {'train': 42, 'validation': 2, 'test': 1}
            else:
                self.fixed_seeds = {}
        else:
            self.fixed_seeds = fixed_seeds.copy()
        super().__init__(space=space)

    def generator(self, part='train'):
        """Yield random ellipse phantom images using
        :meth:`odl.phantom.ellipsoid_phantom`.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.

        Yields
        ------
        image : element of :attr:`space`
            Random ellipse phantom image with values in ``[0., 1.]``.
        """
        seed = self.fixed_seeds.get(part)
        r = np.random.RandomState(seed)
        n_ellipse = 50
        ellipsoids = np.empty((n_ellipse, 6))
        n = self.get_len(part=part)
        it = repeat(None, n) if n is not None else repeat(None)
        for _ in it:
            v = (r.uniform(-.5, .5, (n_ellipse,)) *
                 r.exponential(.4, (n_ellipse,)))
            a1 = .2 * r.exponential(1., (n_ellipse,))
            a2 = .2 * r.exponential(1., (n_ellipse,))
            x = r.uniform(-1., 1., (n_ellipse,))
            y = r.uniform(-1., 1., (n_ellipse,))
            rot = r.uniform(0., 2*np.pi, (n_ellipse,))
            ellipsoids = np.stack((v, a1, a2, x, y, rot), axis=1)
            image = ellipsoid_phantom(self.space, ellipsoids)
            image -= np.min(image)
            image /= np.max(image)

            yield image



# -*- coding: utf-8 -*-
"""Provides wrappers for reconstruction methods of odl."""
from odl.tomo import fbp_op
from odl.operator import Operator
from odl.space.pspace import ProductSpace
from odl.solvers.iterative import iterative, statistical
from dival import Reconstructor, IterativeReconstructor


class FBPReconstructor(Reconstructor):
    HYPER_PARAMS = {
        'filter_type':
            {'default': 'Ram-Lak',
             'choices': ['Ram-Lak', 'Shepp-Logan', 'Cosine', 'Hamming',
                         'Hann']},
        'frequency_scaling':
            {'default': 1.,
             'range': [0, 1],
             'grid_search_options': {'num_samples': 11}}
    }

    """Reconstructor applying filtered back-projection.

    Attributes
    ----------
    fbp_op : `odl.operator.Operator`
        The operator applying filtered back-projection.
    """
    def __init__(self, ray_trafo, padding=True, hyper_params=None,
                 pre_processor=None, post_processor=None, **kwargs):
        """
        Parameters
        ----------
        ray_trafo : `odl.tomo.operators.RayTransform`
            The forward operator. See `odl.tomo.fbp_op` for details.
        padding : bool, optional
            Whether to use padding (the default is ``True``).
            See `odl.tomo.fbp_op` for details.
        pre_processor : callable, optional
            Callable that takes the observation and returns the sinogram that
            is passed to the filtered back-projection operator.
        post_processor : callable, optional
            Callable that takes the filtered back-projection and returns the
            final reconstruction.
        """
        self.ray_trafo = ray_trafo
        self.padding = padding
        self.pre_processor = pre_processor
        self.post_processor = post_processor
        super().__init__(
            reco_space=ray_trafo.domain, observation_space=ray_trafo.range,
            hyper_params=hyper_params, **kwargs)

    def _reconstruct(self, observation, out):
        self.fbp_op = fbp_op(self.ray_trafo, padding=self.padding,
                             **self.hyper_params)
        if self.pre_processor is not None:
            observation = self.pre_processor(observation)
        self.fbp_op(observation, out=out)
        if self.post_processor is not None:
            out[:] = self.post_processor(out)


class CGReconstructor(IterativeReconstructor):
    """Iterative reconstructor applying the conjugate gradient method.

    Attributes
    ----------
    op : `odl.operator.Operator`
        The forward operator of the inverse problem.
    x0 : ``op.domain`` element
        Initial value.
    niter : int
        Number of iterations.
    callback : :class:`odl.solvers.util.callback.Callback` or `None`
        Object that is called in each iteration.
    op_is_symmetric : bool
        If ``False`` (default), the normal equations are solved. If ``True``,
        `op` is assumed to be self-adjoint and the system of equations is
        solved directly.
    """
    def __init__(self, op, x0, niter, callback=None, op_is_symmetric=False,
                 **kwargs):
        """
        Calls `odl.solvers.iterative.iterative.conjugate_gradient_normal` (if
        ``op_is_symmetric == False``) or
        `odl.solvers.iterative.iterative.conjugate_gradient` (if
        ``op_is_symmetric == True``).

        Parameters
        ----------
        op : `odl.operator.Operator`
            The forward operator of the inverse problem.
            If ``op_is_symmetric == False`` (default), the call
            ``op.derivative(x).adjoint`` must be valid.
            If ``op_is_symmetric == True``, `op` must be linear and
            self-adjoint.
        x0 : ``op.domain`` element
            Initial value.
        niter : int
            Number of iterations.
        callback : :class:`odl.solvers.util.callback.Callback`, optional
            Object that is called in each iteration.
        op_is_symmetric : bool, optional
            If ``False`` (default), the normal equations are solved. If
            ``True``, `op` is required to be self-adjoint and the system of
            equations is solved directly.
        """
        self.op = op
        self.x0 = x0
        self.niter = niter
        self.callback = callback
        self.op_is_symmetric = op_is_symmetric
        super().__init__(
            reco_space=self.op.domain, observation_space=self.op.range,
            **kwargs)

    def _reconstruct(self, observation, out):
        observation = self.observation_space.element(observation)
        out[:] = self.x0
        if self.op_is_symmetric:
            iterative.conjugate_gradient(self.op, out, observation,
                                         self.niter, self.callback)
        else:
            iterative.conjugate_gradient_normal(self.op, out, observation,
                                                self.niter, self.callback)
        return out


class GaussNewtonReconstructor(IterativeReconstructor):
    """Iterative reconstructor applying the Gauss-Newton method.

    Attributes
    ----------
    op : `odl.operator.Operator`
        The forward operator of the inverse problem.
    x0 : ``op.domain`` element
        Initial value.
    niter : int
        Maximum number of iterations.
    zero_seq : iterable
        Zero sequence used for regularization.
    callback : :class:`odl.solvers.util.callback.Callback` or `None`
        Object that is called in each iteration.
    """
    def __init__(self, op, x0, niter, zero_seq=None, callback=None, **kwargs):
        """
        Calls `odl.solvers.iterative.iterative.gauss_newton`.

        Parameters
        ----------
        op : `odl.operator.Operator`
            The forward operator of the inverse problem.
            The call ``op.derivative(x).adjoint`` must be valid.
        x0 : ``op.domain`` element
            Initial value.
        niter : int
            Maximum number of iterations.
        zero_seq : iterable, optional
            Zero sequence used for regularization.
        callback : :class:`odl.solvers.util.callback.Callback`, optional
            Object that is called in each iteration.
        """
        self.op = op
        self.x0 = x0
        self.niter = niter
        self.zero_seq = zero_seq
        self.callback = callback
        super().__init__(
            reco_space=self.op.domain, observation_space=self.op.range,
            **kwargs)

    def _reconstruct(self, observation, out):
        observation = self.observation_space.element(observation)
        out[:] = self.x0
        kwargs = {'callback': self.callback}
        if self.zero_seq is not None:
            kwargs['zero_seq'] = self.zero_seq
        iterative.gauss_newton(self.op, out, observation, self.niter,
                               **kwargs)
        return out


class KaczmarzReconstructor(IterativeReconstructor):
    """Iterative reconstructor applying Kaczmarz's method.

    Attributes
    ----------
    ops : sequence of `odl.Operator`
        The forward operators of the inverse problem.
    x0 : ``op.domain`` element
        Initial value.
    niter : int
        Number of iterations.
    omega : positive float or sequence of positive floats
        Relaxation parameter.
        If a single float is given it is used for all operators.
    random : bool
        Whether the order of the operators is randomized in each iteration.
    projection : callable
        Callable that can be used to modify the iterates in each iteration.
    callback : :class:`odl.solvers.util.callback.Callback` or `None`
        Object that is called in each iteration.
    callback_loop : {'inner', 'outer'}
        Whether the `callback` should be called in the inner or outer loop.
    """
    def __init__(self, ops, x0, niter, omega=1, random=False, projection=None,
                 callback=None, callback_loop='outer', **kwargs):
        """
        Calls `odl.solvers.iterative.iterative.kaczmarz`.

        Parameters
        ----------
        ops : sequence of `odl.Operator`
            The forward operators of the inverse problem.
            The call ``ops[i].derivative(x).adjoint`` must be valid for all i.
        x0 : ``op.domain`` element
            Initial value.
        niter : int
            Number of iterations.
        omega : positive float or sequence of positive floats, optional
            Relaxation parameter.
            If a single float is given it is used for all operators.
        random : bool, optional
            Whether the order of the operators is randomized in each iteration.
        projection : callable, optional
            Callable that can be used to modify the iterates in each iteration.
        callback : :class:`odl.solvers.util.callback.Callback`, optional
            Object that is called in each iteration.
        callback_loop : {'inner', 'outer'}
            Whether the `callback` should be called in the inner or outer loop.
        """
        self.ops = ops
        self.x0 = x0
        self.niter = niter
        self.omega = omega
        self.projection = projection
        self.random = random
        self.callback = callback
        self.callback_loop = callback_loop
        super().__init__(
            reco_space=self.ops[0].domain,
            observation_space=ProductSpace(*(op.range for op in self.ops)),
            **kwargs)

    def _reconstruct(self, observation, out):
        observation = self.observation_space.element(observation)
        out[:] = self.x0
        iterative.kaczmarz(self.ops, out, observation, self.niter,
                           self.omega, self.projection, self.random,
                           self.callback, self.callback_loop)
        return out


class LandweberReconstructor(IterativeReconstructor):
    """Iterative reconstructor applying Landweber's method.

    Attributes
    ----------
    op : `odl.operator.Operator`
        The forward operator of the inverse problem.
    x0 : ``op.domain`` element
        Initial value.
    niter : int
        Number of iterations.
    omega : positive float
        Relaxation parameter.
    projection : callable
        Callable that can be used to modify the iterates in each iteration.
    callback : :class:`odl.solvers.util.callback.Callback` or `None`
        Object that is called in each iteration.
    """
    def __init__(self, op, x0, niter, omega=None, projection=None,
                 callback=None, **kwargs):
        """
        Calls `odl.solvers.iterative.iterative.landweber`.

        Parameters
        ----------
        op : `odl.operator.Operator`
            The forward operator of the inverse problem.
            The call ``op.derivative(x).adjoint`` must be valid.
        x0 : ``op.domain`` element
            Initial value.
        niter : int
            Number of iterations.
        omega : positive float, optional
            Relaxation parameter.
        projection : callable, optional
            Callable that can be used to modify the iterates in each iteration.
            One argument is passed and expected be modified in-place.
        callback : :class:`odl.solvers.util.callback.Callback`, optional
            Object that is called in each iteration.
        """
        self.op = op
        self.x0 = x0
        self.niter = niter
        self.omega = omega
        self.projection = projection
        self.callback = callback
        super().__init__(
            reco_space=self.op.domain, observation_space=self.op.range,
            **kwargs)

    def _reconstruct(self, observation, out):
        observation = self.observation_space.element(observation)
        out[:] = self.x0
        iterative.landweber(self.op, out, observation, self.niter,
                            self.omega, self.projection, self.callback)
        return out


class MLEMReconstructor(IterativeReconstructor):
    """Iterative reconstructor applying Maximum Likelihood Expectation
    Maximization.

    If multiple operators are given, Ordered Subsets MLEM is applied.

    Attributes
    ----------
    op : `odl.operator.Operator` or sequence of `odl.operator.Operator`
        The forward operator(s) of the inverse problem.
    x0 : ``op.domain`` element
        Initial value.
    niter : int
        Number of iterations.
    noise : {'poisson'} or `None`
        Noise model determining the variant of MLEM.
    callback : :class:`odl.solvers.util.callback.Callback` or `None`
        Object that is called in each iteration.
    sensitivities : float or ``op.domain`` `element-like`
        Usable with ``noise='poisson'``. The algorithm contains an ``A^T 1``
        term, if this parameter is given, it is replaced by it.
        Default: ``op[i].adjoint(op[i].range.one())``
    """
    def __init__(self, op, x0, niter, noise='poisson', callback=None,
                 sensitivities=None, **kwargs):
        """
        Calls `odl.solvers.iterative.statistical.osmlem`.

        Parameters
        ----------
        op : `odl.operator.Operator` or sequence of `odl.operator.Operator`
            The forward operator(s) of the inverse problem.
            If an operator sequence is given, Ordered Subsets MLEM is applied.
        x0 : ``op.domain`` element
            Initial value.
        niter : int
            Number of iterations.
        noise : {'poisson'}, optional
            Noise model determining the variant of MLEM.
            For ``'poisson'``, the initial value of ``x`` should be
            non-negative.
        callback : :class:`odl.solvers.util.callback.Callback`, optional
            Object that is called in each iteration.
        sensitivities : float or ``op.domain`` `element-like`, optional
            Usable with ``noise='poisson'``. The algorithm contains an
            ``A^T 1`` term, if this parameter is given, it is replaced by it.
            Default: ``op[i].adjoint(op[i].range.one())``
        """
        self.os_mode = not isinstance(op, Operator)
        self.op = op if self.os_mode else [op]
        self.x0 = x0
        self.niter = niter
        self.noise = noise
        self.callback = callback
        self.sensitivities = sensitivities
        observation_space = (ProductSpace(*(op.range for op in self.op)) if
                             self.os_mode else self.op[0].range)
        super().__init__(
            reco_space=self.op[0].domain, observation_space=observation_space,
            **kwargs)

    def _reconstruct(self, observation, out):
        out[:] = self.x0
        observation = self.observation_space.element(observation)
        if not self.os_mode:
            observation = [observation]
        statistical.osmlem(self.op, out, observation, self.niter,
                           noise=self.noise, callback=self.callback,
                           sensitivities=self.sensitivities)
        return out



# -*- coding: utf-8 -*-
"""
Provides the abstract :class:`Measure` base class and some popular measures.

Measure instances are identified by a unique :attr:`~Measure.short_name`, which
is used by the :mod:`~dival.evaluation` module.
"""
from abc import ABC, abstractmethod
from warnings import warn
import numpy as np
from skimage.measure import compare_ssim
from odl.operator.operator import Operator


def gen_unique_name(name_orig):
    i = 1
    while True:
        yield '{}_{}'.format(name_orig, str(i))
        i += 1


class Measure(ABC):
    """Abstract base class for measures used for evaluation.

    In subclasses, either :meth:`__init__` should be inherited or it should
    call ``super().__init__()`` in order to register the :attr:`short_name`
    and to ensure it is unique.

    Attributes
    ----------
    measure_type : {``'distance'``, ``'quality'``}
        The measure type.
        Measures with type ``'distance'`` should attain small values if the
        reconstruction is good. Measures with type ``'quality'`` should attain
        large values if the reconstruction is good.
    short_name : str
        Short name of the measure, used as identifier (key in
        :attr:`measure_dict`).
    name : str
        Name of the measure.
    description : str
        Description of the measure.
    measure_dict : dict, class attribute
        Registry of all measures with their :attr:`short_name` as key.
    """
    measure_type = None
    short_name = ''
    name = ''
    description = ''

    measure_dict = {}

    def __init__(self, short_name=None):
        """
        Parameters
        ----------
        short_name : str, optional
            The short name of this measure, used as identifier in
            :attr:`measure_dict`.
            If `None` is passed and :attr:`short_name` was not set by the
            subclass, the ``__name__`` of the subclass is used.
            If `short_name` is already taken by another instance, a unique
            short name is generated by appending a suffix of format ``'_%d'``.
        """
        if short_name is not None:
            self.short_name = short_name
        elif self.short_name is None:
            self.short_name = self.__class__.__name__
        if self.short_name in self.__class__.measure_dict:
            old_short_name = self.short_name
            unique_name = gen_unique_name(self.short_name)
            while self.short_name in self.__class__.measure_dict:
                self.short_name = next(unique_name)
            warn("Measure `short_name` '{}' already exists, changed to '{}'"
                 .format(old_short_name, self.short_name))
        self.__class__.measure_dict[self.short_name] = self

    @abstractmethod
    def apply(self, reconstruction, ground_truth):
        """Calculate the value of this measure.

        Parameters
        ----------
        reconstruction : odl element
            The reconstruction.
        ground_truth : odl element
            The ground truth to compare with.

        Returns
        -------
        value : float
            The value of this measure for the given `reconstruction` and
            `ground_truth`.
        """

    def __call__(self, reconstruction, ground_truth):
        """Call :meth:`apply`.
        """
        return self.apply(self, reconstruction, ground_truth)

    @classmethod
    def get_by_short_name(cls, short_name):
        """
        Return :class:`.Measure` instance with given short name by registry
        lookup.

        Parameters
        ----------
        short_name : str
            Short name, identifier in :attr:`measure_dict`.

        Returns
        -------
        measure : :class:`.Measure`
            The instance.
        """
        return cls.measure_dict.get(short_name)

    class _OperatorForFixedGroundTruth(Operator):
        def __init__(self, measure, ground_truth):
            super().__init__(ground_truth.space, ground_truth.space.field)
            self.measure = measure
            self.ground_truth = ground_truth

        def _call(self, x):
            return self.measure.apply(x, self.ground_truth)

    def as_operator_for_fixed_ground_truth(self, ground_truth):
        """
        Return an odl operator that can be applied to different reconstructions
        for fixed ground truth.

        Returns
        -------
        op : odl operator
            odl operator.
        """
        return self._OperatorForFixedGroundTruth(self, ground_truth)


class L2Measure(Measure):
    """The euclidean (l2) distance measure."""
    measure_type = 'distance'
    short_name = 'l2'
    name = 'euclidean distance'
    description = ('distance given by '
                   'sqrt(sum((reconstruction-ground_truth)**2))')

    def apply(self, reconstruction, ground_truth):
        return np.linalg.norm((np.asarray(reconstruction) -
                               np.asarray(ground_truth)).flat)


L2 = L2Measure()


class MSEMeasure(Measure):
    """The mean squared error distance measure."""
    measure_type = 'distance'
    short_name = 'mse'
    name = 'mean squared error'
    description = ('distance given by '
                   '1/n * sum((reconstruction-ground_truth)**2)')

    def apply(self, reconstruction, ground_truth):
        return np.mean((np.asarray(reconstruction) -
                        np.asarray(ground_truth))**2)


MSE = MSEMeasure()


class PSNRMeasure(Measure):
    """The peak signal-to-noise ratio (PSNR) measure.

    The data range is automatically determined from the ground truth if not
    given to the constructor.

    Attributes
    ----------
    data_range : float or `None`
        The data range (max-min possible value).
        If `data_range` is `None`,
        ``np.max(ground_truth) - np.min(ground_truth)`` is used in
        :meth:`apply`.
    """
    measure_type = 'quality'
    short_name = 'psnr'
    name = 'peak signal-to-noise ratio'
    description = 'quality given by 10*log10(MAX**2/MSE)'

    def __init__(self, data_range=None, short_name=None):
        """
        Parameters
        ----------
        data_range : float, optional
            The data range (max-min possible value).
            If `data_range` is `None`,
            ``np.max(ground_truth) - np.min(ground_truth)`` is used in
            :meth:`apply`.
        short_name : str, optional
            Short name.
        """
        self.data_range = data_range
        if self.data_range is not None and short_name is None:
            short_name = '{}_data_range{}'.format(self.__class__.short_name,
                                                  self.data_range)
        super().__init__(short_name=short_name)

    def apply(self, reconstruction, ground_truth):
        gt = np.asarray(ground_truth)
        mse = np.mean((np.asarray(reconstruction) - gt)**2)
        if mse == 0.:
            return float('inf')
        data_range = self.data_range or (np.max(gt) - np.min(gt))
        return 20*np.log10(data_range) - 10*np.log10(mse)


PSNR = PSNRMeasure()


class SSIMMeasure(Measure):
    """The structural similarity index measure."""
    measure_type = 'quality'
    short_name = 'ssim'
    name = 'structural similarity index'
    description = ('The (M)SSIM like described in `Wang et al. 2014 '
                   '<https://doi.org/10.1109/TIP.2003.819861>`_.')

    def __init__(self, short_name=None, **kwargs):
        """
        This is a wrapper for :func:`skimage.measure.compare_ssim`.
        The data range is automatically determined from the ground truth if not
        given to the constructor.

        Parameters
        ----------
        short_name : str, optional
            Short name.
        kwargs : dict, optional
            Keyword arguments that will be passed to
            :func:`~skimage.measure.compare_ssim` in :meth:`apply`.
            If `data_range` is not specified,
            ``np.max(ground_truth) - np.min(ground_truth)`` is used.
        """
        self.kwargs = kwargs
        self.data_range = self.kwargs.pop('data_range', None)
        if self.data_range is not None and short_name is None:
            short_name = '{}_data_range{}'.format(self.__class__.short_name,
                                                  self.data_range)
        super().__init__(short_name=short_name)

    def apply(self, reconstruction, ground_truth):
        gt = np.asarray(ground_truth)
        data_range = (self.data_range if self.data_range is not None
                      else np.max(gt) - np.min(gt))
        return compare_ssim(reconstruction, gt, data_range=data_range,
                            **self.kwargs)


SSIM = SSIMMeasure()



import numpy as np
import odl
from dival.data import DataPairs
from dival.evaluation import TaskTable
from dival.measure import PSNR, SSIM
from dival.reconstructors.odl_reconstructors import (FBPReconstructor,
                                                     CGReconstructor,
                                                     GaussNewtonReconstructor,
                                                     LandweberReconstructor,
                                                     MLEMReconstructor)

np.random.seed(0)

# %% data
reco_space = odl.uniform_discr(
    min_pt=[-20, -20], max_pt=[20, 20], shape=[300, 300],
    dtype='float32')
phantom = odl.phantom.shepp_logan(reco_space, modified=True)
ground_truth = phantom

geometry = odl.tomo.cone_beam_geometry(reco_space, 40, 40, 360)
ray_trafo = odl.tomo.RayTransform(reco_space, geometry, impl='astra_cpu')
proj_data = ray_trafo(phantom)
observation = (proj_data + np.random.poisson(0.3, proj_data.shape)).asarray()

test_data = DataPairs(observation, ground_truth, name='shepp-logan + pois')

# %% task table and reconstructors
eval_tt = TaskTable()

fbp_reconstructor = FBPReconstructor(ray_trafo, hyper_params={
    'filter_type': 'Hann',
    'frequency_scaling': 0.8})
cg_reconstructor = CGReconstructor(ray_trafo, reco_space.zero(), 4)
gn_reconstructor = GaussNewtonReconstructor(ray_trafo, reco_space.zero(), 2)
lw_reconstructor = LandweberReconstructor(ray_trafo, reco_space.zero(), 8)
mlem_reconstructor = MLEMReconstructor(ray_trafo, 0.5*reco_space.one(), 1)

reconstructors = [fbp_reconstructor, cg_reconstructor, gn_reconstructor,
                  lw_reconstructor, mlem_reconstructor]
options = {'save_iterates': True}

eval_tt.append_all_combinations(reconstructors=reconstructors,
                                test_data=[test_data], options=options)

# %% run task table
results = eval_tt.run()
results.apply_measures([PSNR, SSIM])
print(results)

# %% plot reconstructions
fig = results.plot_all_reconstructions(fig_size=(9, 4), vrange='individual')

# %% plot convergence
results.plot_convergence(1, fig_size=(9, 6), gridspec_kw={'hspace': 0.5})
results.plot_performance(PSNR)



import numpy as np
from hyperopt import hp
from odl.discr import uniform_discr
from dival.datasets.dataset import Dataset
from dival.evaluation import TaskTable
from dival.measure import L2, PSNR
from dival.reconstructors.regression_reconstructors import LinRegReconstructor
from hyper_param_search import optimize_hyper_params

# %% data
observation_space = uniform_discr(-0.5, 6.5, 7)
reco_space = uniform_discr(-0.5, 11.5, 12)

np.random.seed(0)


class LinearDataset(Dataset):
    def __init__(self, observation_space, reco_space, train_len=10000,
                 validation_len=1000, test_len=1000):
        self.train_len = train_len
        self.validation_len = validation_len
        self.test_len = test_len
        self.observation_space = observation_space
        self.reco_space = reco_space
        self.shape = (self.observation_space.shape, self.reco_space.shape)
        self.forward_matrix = np.random.rand(self.shape[0][0],
                                             self.shape[1][0])

    def generator(self, part='train'):
        seed = 42
        if part == 'validation':
            seed = 2
        elif part == 'test':
            seed = 1
        rs = np.random.RandomState(seed)
        for _ in range(self.get_len(part=part)):
            x = rs.rand(self.shape[1][0])
            y = (np.dot(self.forward_matrix, x) +
                 0.4 * rs.normal(scale=.1, size=self.shape[0]))
            yield (self.observation_space.element(y),
                   self.reco_space.element(x))


dataset = LinearDataset(observation_space, reco_space)
validation_data = dataset.get_data_pairs('validation', 10)
test_data = dataset.get_data_pairs('test', 10)

# %% task table and reconstructors
eval_tt = TaskTable()

reconstructor = LinRegReconstructor(observation_space=observation_space,
                                    reco_space=reco_space)
optimize_hyper_params(reconstructor,
                      validation_data=validation_data,
                      measure=L2,
                      dataset=dataset,
                      hyperopt_max_evals_retrain=100,
                      HYPER_PARAMS_override={
                          'l2_regularization': {
                              'method': 'hyperopt',
                              'hyperopt_options': {
                                  'space': hp.loguniform('l2_regularization',
                                                         0., np.log(1e9))
                              }
                          }})
print('optimized l2 reg. coeff.: {}'.format(
    reconstructor.hyper_params['l2_regularization']))

eval_tt.append(reconstructor=reconstructor, test_data=test_data,
               dataset=dataset, measures=[L2, PSNR])

# %% run task table
results = eval_tt.run()
print(results)

# %% plot reconstructions
fig = results.plot_all_reconstructions(test_index=range(3),
                                       fig_size=(9, 4), vrange='individual')



# -*- coding: utf-8 -*-
__version__ = '0.3.1'

from .config import CONFIG, get_config, set_config
from .data import DataPairs
from .datasets import Dataset
from .datasets.standard import get_standard_dataset
from .reconstructors import (Reconstructor, IterativeReconstructor,
                             LearnedReconstructor)
from .measure import Measure
from .evaluation import TaskTable


__all__ = ['CONFIG', 'get_config', 'set_config',
           'DataPairs',
           'Dataset', 'get_standard_dataset',
           'Reconstructor', 'IterativeReconstructor', 'LearnedReconstructor',
           'Measure',
           'TaskTable']



# -*- coding: utf-8 -*-
import os
import requests
import tqdm
import hashlib


def download_file(url, filename=False, chunk_size=1024, verbose=False,
                  md5sum=True):
    """
    Download file with progressbar.

    Based on
    `<https://gist.github.com/ruxi/5d6803c116ec1130d484a4ab8c00c603>`_
    (MIT License).

    Parameters
    ----------
    url : str
        URL to download from.
    filename : str, optional
        Local destination filename (including path).
        By default, the file is stored under the current folder and the name
        given by the last part of `url`.
    chunk_size : int, optional
        Number of bytes in a chunk.
    verbose : bool, optional
        Whether to print additional information.
    md5sum : bool, optional
        Whether to compute and return the MD5 checksum (hex-digest).

    Returns
    -------
    md5sum : str or `None`
        Hex-digest of the MD5 hash, if ``md5sum=True``, otherwise `None`.
    """
    if not filename:
        local_filename = os.path.join(".", url.split('/')[-1])
    else:
        local_filename = filename
    r = requests.get(url, stream=True)
    file_size = int(r.headers['Content-Length'])
    chunk = 1
    num_bars = int(file_size / chunk_size)
    if verbose:
        print(dict(file_size=file_size))
        print(dict(num_bars=num_bars))

    if md5sum:
        hash_md5 = hashlib.md5()

    with open(local_filename, 'wb') as fp:
        for chunk in tqdm.tqdm(r.iter_content(chunk_size=chunk_size),
                               total=num_bars,
                               unit='chunks',
                               desc=local_filename,
                               leave=True):
            fp.write(chunk)
            if md5sum:
                hash_md5.update(chunk)
    if md5sum:
        return hash_md5.hexdigest()


def input_yes_no(default='y'):
    """
    Demand user input y[es] or n[o].

    The user is asked repeatedly, until the input is valid.

    Parameters
    ----------
    default : {``'y'``, ``'n'``}, optional
        The output if the user enters empty input.

    Returns
    -------
    inp : {``'y'``, ``'n'``}
        The users input (or `default`).
    """
    def _input():
        inp = input()
        inp = inp.lower()
        if inp in ['y', 'yes']:
            inp = 'y'
        elif inp in ['n', 'no']:
            inp = 'n'
        elif inp == '':
            inp = default
        else:
            print('please input y[es] or n[o]')
            return None
        return inp

    inp = _input()
    while inp not in ['y', 'n']:
        inp = _input()

    return inp == 'y'


def download_zenodo_record(record_id, base_path='', md5sum_check=True):
    """
    Download a zenodo record.

    Parameters
    ----------
    record_id : str
        Record id.
    base_path : str, optional
        Path to store the downloaded files in. Default is the current folder.
    md5sum_check : bool, optional
        Whether to check the MD5 sum of each downloaded file.

    Returns
    -------
    success : bool
        If ``md5sum_check=True``, whether all sums matched. Otherwise the
        returned value is always ``True``.
    """
    r = requests.get('https://zenodo.org/api/records/{}'.format(record_id))
    files = r.json()['files']
    success = True
    for i, f in enumerate(files):
        url = f['links']['self']
        filename = f['key']
        size_kb = f['size'] / 1000
        checksum = f['checksum']
        print("downloading file {:d}/{:d}: '{}', {}KB".format(
            i+1, len(files), filename, size_kb))
        if md5sum_check:
            retry = True
            md5sum_matches = False
            while retry and not md5sum_matches:
                md5sum = download_file(url, os.path.join(base_path, filename),
                                       md5sum=True)
                md5sum_matches = (md5sum == checksum.split(':')[1])
                if not md5sum_matches:
                    print("md5 checksum does not match for file '{}'. Retry "
                          "downloading? (y)/n".format(filename))
                    retry = input_yes_no()
            if not md5sum_matches:
                success = False
                print('record download aborted')
                break
        else:
            download_file(url, os.path.join(base_path, filename), md5sum=False)
    return success



# -*- coding: utf-8 -*-
"""Redirecting writing to ``tqdm`` (the progressbar).

Taken from `tqdm docs <https://github.com/tqdm/tqdm#redirecting-writing>`_.
"""
import contextlib
import sys
from tqdm import tqdm


class DummyTqdmFile(object):
    """Dummy file-like that will write to ``tqdm``.

    Attributes
    ----------
    file : file object
        File to write to using :func:`tqdm.write`.
    """
    file = None

    def __init__(self, file):
        """
        Parameters
        ----------
        file : file object
            File to write to using :func:`tqdm.write`.
        """
        self.file = file

    def write(self, x):
        """
        Call ``tqdm.write``.

        Parameters
        ----------
        x : str
            Text.
        """
        # Avoid print() second call (useless \n)
        if len(x.rstrip()) > 0:
            tqdm.write(x, file=self.file)

    def flush(self):
        """
        Flush :attr:`file`.
        """
        return getattr(self.file, "flush", lambda: None)()


@contextlib.contextmanager
def std_out_err_redirect_tqdm(std_out=None):
    """
    Context manager that redirects :obj:`sys.stdout` and :obj:`sys.stderr` to
    ``tqdm``.

    std_out : `None` or file object
        If `None` is passed, :obj:`sys.stdout` and :obj:`sys.stderr` are
        redirected to ``tqdm``.
        If a file object is passed, it is yielded immediately and no
        redirection is done. This is useful for nested calls:

        .. code-block:: python

            def func(std_out=None)
                with std_out_err_redirect_tqdm(std_out) as std_out:
                    for i in tqdm(range(7), file=std_out):
                        print(i)

            with std_out_err_redirect_tqdm() as std_out:
                func(std_out=std_out)

    Yields
    ------
    std_out : file object
        The original :obj:`sys.stdout` if ``std_out=None`` is passed, else
        `std_out`.
    """
    if std_out is not None:
        yield std_out
    else:
        orig_out_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = map(DummyTqdmFile, orig_out_err)
            yield orig_out_err[0]
        # Relay exceptions
        except Exception as exc:
            raise exc
        # Always restore sys.stdout/err if necessary
        finally:
            sys.stdout, sys.stderr = orig_out_err



# -*- coding: utf-8 -*-
from .reconstructor import (Reconstructor, IterativeReconstructor,
                            LearnedReconstructor, FunctionReconstructor)

__all__ = ['Reconstructor', 'IterativeReconstructor', 'LearnedReconstructor',
           'FunctionReconstructor']



# -*- coding: utf-8 -*-
"""
Provides the learned :class:`FBPUNetReconstructor` for the application of CT.
"""
import copy
from warnings import warn
import numpy as np
from odl.tomo import fbp_op
from tqdm import tqdm
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from dival import Dataset, LearnedReconstructor
from skimage.measure import compare_psnr


def generate_fbp_cache(dataset, part, filename, ray_trafo, size=None):
    """
    Write filtered back-projections for a CT dataset part to file.

    Parameters
    ----------
    dataset : :class:`.Dataset`
        CT dataset from which the observations are used.
    part : {``'train'``, ``'validation'``, ``'test'``}
        The data part.
    filename : str
        The filename to store the FBP cache at (ending ``.npy``).
    ray_trafo : :class:`odl.tomo.RayTransform`
        Ray transform from which the FBP operator is constructed.
    size : int, optional
        Number of samples to use from the dataset.
        By default, all samples are used.
    """
    fbp = fbp_op(ray_trafo)
    num_samples = dataset.get_len(part=part) if size is None else size
    reco_fbps = np.empty((num_samples,) + dataset.shape[1], dtype=np.float32)
    obs = np.empty(dataset.shape[0], dtype=np.float32)
    tmp_fbp = fbp.range.element()
    for i in tqdm(range(num_samples), desc='generating FBP cache'):
        dataset.get_sample(i, part=part, out=(obs, False))
        fbp(obs, out=tmp_fbp)
        reco_fbps[i][:] = tmp_fbp
    np.save(filename, reco_fbps)


class CachedFBPDataset(Dataset):
    """Dataset combining the ground truth of a dataset with cached FBPs.

    Each sample is a pair of a FBP and a ground truth image.
    """
    def __init__(self, dataset, filenames):
        """
        Parameters
        ----------
        dataset : :class:`.Dataset`
            CT dataset from which the ground truth is used.
        filenames : dict
            Cached FBP filenames for the dataset parts.
            The part (``'train'``, ...) is the key to the dict.
            To generate the FBP files, :func:`generate_fbp_cache` can be used.
        """
        self.dataset = dataset
        self.filenames = filenames
        self.fbps = {}
        for part, filename in filenames.items():
            try:
                self.fbps[part] = np.load(filename, mmap_mode='r')
            except FileNotFoundError:
                raise FileNotFoundError(
                    "Did not find FBP cache file '{}'. Please specify an "
                    'existing file or generate it using '
                    '`fbp_unet_reconstructor.generate_fbp_cache(...)`.'.format(
                        filename))
        self.train_len = (len(self.fbps['train']) if 'train' in self.fbps else
                          self.dataset.get_len('train'))
        self.validation_len = (len(self.fbps['validation']) if 'validation' in
                               self.fbps else
                               self.dataset.get_len('validation'))
        self.test_len = (len(self.fbps['test']) if 'test' in self.fbps else
                         self.dataset.get_len('test'))
        self.shape = (self.dataset.shape[1], self.dataset.shape[1])
        self.num_elements_per_sample = 2
        self.random_access = True
        super().__init__(space=(self.dataset.space[1], self.dataset.space[1]))

    def get_sample(self, index, part='train', out=None):
        if out is None:
            out = (True, True)
        (_, gt) = self.dataset.get_sample(index, part=part,
                                          out=(False, out[1]))
        if isinstance(out[0], bool):
            fbp = self.fbps[part][index] if out[0] else None
        else:
            out[0][:] = self.fbps[part][index]
            fbp = out[0]
        return (fbp, gt)

    def get_samples(self, key, part='train', out=None):
        if out is None:
            out = (True, True)
        (_, gt) = self.dataset.get_samples(key, part=part,
                                           out=(False, out[1]))
        if isinstance(out[0], bool):
            fbp = self.fbps[part][key] if out[0] else None
        else:
            out[0][:] = self.fbps[part][key]
            fbp = out[0]
        return (fbp, gt)


class FBPDataset(Dataset):
    """
    Dataset computing filtered back-projections for a CT dataset on the fly.

    Each sample is a pair of a FBP and a ground truth image.
    """
    def __init__(self, dataset, ray_trafo):
        """
        Parameters
        ----------
        dataset : :class:`.Dataset`
            CT dataset. FBPs are computed from the observations, the ground
            truth is taken directly from the dataset.
        ray_trafo : :class:`odl.tomo.RayTransform`
            Ray transform from which the FBP operator is constructed.
        """
        self.dataset = dataset
        self.fbp_op = fbp_op(ray_trafo)
        self.train_len = self.dataset.get_len('train')
        self.validation_len = self.dataset.get_len('validation')
        self.test_len = self.dataset.get_len('test')
        self.shape = (self.dataset.shape[1], self.dataset.shape[1])
        self.num_elements_per_sample = 2
        self.random_access = dataset.supports_random_access()
        super().__init__(space=(self.dataset.space[1], self.dataset.space[1]))

    def generator(self, part='train'):
        gen = self.dataset.generator(part=part)
        for (obs, gt) in gen:
            fbp = self.fbp_op(obs)
            yield (fbp, gt)

    def get_sample(self, index, part='train', out=None):
        if out is None:
            out = (True, True)
        out_fbp = not (isinstance(out[0], bool) and not out[0])
        (obs, gt) = self.dataset.get_sample(index, part=part,
                                            out=(out_fbp, out[1]))
        if isinstance(out[0], bool):
            fbp = self.fbp_op(obs) if out[0] else None
        else:
            if out[0] in self.fbp_op.range:
                self.fbp_op(obs, out=out[0])
            else:
                out[0][:] = self.fbp_op(obs)
            fbp = out[0]
        return (fbp, gt)

    def get_samples(self, key, part='train', out=None):
        if out is None:
            out = (True, True)
        out_fbp = not (isinstance(out[0], bool) and not out[0])
        (obs_arr, gt_arr) = self.dataset.get_samples(key, part=part,
                                                     out=(out_fbp, out[1]))
        if isinstance(out[0], bool) and out[0]:
            fbp_arr = np.empty((len(obs_arr),) + self.dataset.shape[1],
                               dtype=self.dataset.space[1].dtype)
        elif isinstance(out[0], bool) and not out[0]:
            fbp_arr = None
        else:
            fbp_arr = out[0]
        if out_fbp:
            tmp_fbp = self.fbp_op.range.element()
            for i in range(len(obs_arr)):
                self.fbp_op(obs_arr[i], out=tmp_fbp)
                fbp_arr[i][:] = tmp_fbp
        return (fbp_arr, gt_arr)


class FBPUNetReconstructor(LearnedReconstructor):
    """
    CT Reconstructor applying filtered back-projection followed by a
    postprocessing U-Net (cf. [1]_).
    The U-Net architecture is similar to the one in [2]_.

    References
    ----------
    .. [1] K. H. Jin, M. T. McCann, E. Froustey, et al., 2017,
           "Deep Convolutional Neural Network for Inverse Problems in Imaging".
           IEEE Transactions on Image Processing.
           `doi:10.1109/TIP.2017.2713099
           <https://doi.org/10.1109/TIP.2017.2713099>`_
    .. [2] V. Lempitsky, A. Vedaldi, and D. Ulyanov, 2018,
           "Deep Image Prior".
           IEEE/CVF Conference on Computer Vision and Pattern Recognition.
           `doi:10.1109/CVPR.2018.00984
           <https://doi.org/10.1109/CVPR.2018.00984>`_
    """
    def __init__(self, ray_trafo, scales=5, epochs=20, batch_size=64,
                 num_data_loader_workers=8, use_cuda=True, show_pbar=True,
                 fbp_impl='astra_cuda', **kwargs):
        """
        Parameters
        ----------
        ray_trafo : :class:`odl.tomo.RayTransform`
            Ray transform from which the FBP operator is constructed.
        scales : int
            Number of scales in the U-Net.
        epochs : int
            Number of epochs to train.
        batch_size : int
            Batch size.
        num_data_loader_workers : int
            Number of parallel workers to use for loading data.
        use_cuda : bool
            Whether to use cuda for the U-Net.
        show_pbar : bool
            Whether to show tqdm progress bars during the epochs.
        fbp_impl : str, optional
            The backend implementation passed to
            :class:`odl.tomo.RayTransform` in case no `ray_trafo` is specified.
            Then ``dataset.get_ray_trafo(impl=fbp_impl)`` is used to get the
            ray transform and FBP operator.
        """
        self.ray_trafo = ray_trafo
        self.fbp_op = fbp_op(self.ray_trafo)
        self.epochs = epochs
        self.batch_size = batch_size
        self.num_data_loader_workers = num_data_loader_workers
        self.use_cuda = use_cuda
        self.show_pbar = show_pbar
        self.fbp_impl = fbp_impl
        tensor_type = (torch.cuda.FloatTensor if self.use_cuda else
                       torch.FloatTensor)
        self.net = get_skip_model(scales=scales).type(tensor_type)
        super().__init__(reco_space=self.ray_trafo.domain,
                         observation_space=self.ray_trafo.range, **kwargs)

    def train(self, dataset):
        try:
            self.fbp_dataset = dataset.fbp_dataset
        except AttributeError:
            self.fbp_dataset = FBPDataset(dataset, self.ray_trafo)

        num_workers = self.num_data_loader_workers
        if not self.fbp_dataset.supports_random_access():
            warn('Dataset does not support random access. Shuffling will not '
                 'work, and only 1 worker will be used for data loading.')
            num_workers = 1
        fbp_dataset_train = self.fbp_dataset.create_torch_dataset(
            part='train', reshape=((1,) + self.fbp_dataset.shape[0],
                                   (1,) + self.fbp_dataset.shape[1]))
        fbp_dataset_validation = self.fbp_dataset.create_torch_dataset(
            part='validation', reshape=((1,) + self.fbp_dataset.shape[0],
                                        (1,) + self.fbp_dataset.shape[1]))

        criterion = torch.nn.MSELoss()
        optimizer = torch.optim.Adam(self.net.parameters(), lr=0.01)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=5,
                                                    gamma=0.1)

        dataloaders = {'train': DataLoader(
                           fbp_dataset_train, batch_size=self.batch_size,
                           num_workers=num_workers, shuffle=True,
                           pin_memory=True),
                       'validation': DataLoader(
                           fbp_dataset_validation, batch_size=self.batch_size,
                           num_workers=num_workers,
                           shuffle=True, pin_memory=True)}

        dataset_sizes = {'train': len(fbp_dataset_train),
                         'validation': len(fbp_dataset_validation)}

        best_model_wts = copy.deepcopy(self.net.state_dict())
        best_psnr = 0

        self.net = nn.DataParallel(self.net)

        device = (torch.device('cuda:0') if self.use_cuda else
                  torch.device('cpu'))
        self.net.to(device)
        self.net.train()

        self.history = {'train': {'loss': [],
                                  'psnr': []},
                        'validation': {'loss': [],
                                       'psnr': []}}

        for epoch in range(self.epochs):
            print('epoch {:d}/{:d}'.format(epoch, self.epochs), flush=True)
            # Each epoch has a training and validation phase
            for phase in ['train', 'validation']:
                if phase == 'train':
                    self.net.train()  # Set model to training mode
                else:
                    self.net.eval()   # Set model to evaluate mode

                running_psnr = 0.0
                # Iterate over data.
                batch_id = 0
                with tqdm(dataloaders[phase], desc='epoch {:d}'.format(epoch),
                          disable=not self.show_pbar) as pbar:
                    for inputs, labels in pbar:
                        inputs = inputs.to(device)
                        labels = labels.to(device)

                        # zero the parameter gradients
                        optimizer.zero_grad()

                        # forward
                        # track gradients only if in train phase
                        with torch.set_grad_enabled(phase == 'train'):
                            outputs = self.net(inputs)
                            loss = criterion(outputs, labels)

                            # backward + optimize only if in training phase
                            if phase == 'train':
                                loss.backward()
                                optimizer.step()

                        # statistics
                        current_loss = loss.item()
                        current_psnr = 0.0
                        for i in range(outputs.shape[0]):
                            labels_ = labels[i, 0].detach().cpu().numpy()
                            outputs_ = outputs[i, 0].detach().cpu().numpy()
                            current_psnr += compare_psnr(
                                outputs_, labels_, data_range=np.max(labels_))
                        current_psnr /= outputs.shape[0]
                        running_psnr += current_psnr
                        self.history[phase]['loss'].append(current_loss)
                        self.history[phase]['psnr'].append(current_psnr)
                        pbar.set_postfix({'phase': phase,
                                          'loss': current_loss,
                                          'psnr': current_psnr})

                        batch_id += 1

                    epoch_psnr = running_psnr / dataset_sizes[phase]

                    # deep copy the model
                    if phase == 'validation' and epoch_psnr > best_psnr:
                        best_psnr = epoch_psnr
                        best_model_wts = copy.deepcopy(self.net.state_dict())
                scheduler.step(epoch=epoch)

        print('Best val psnr: {:4f}'.format(best_psnr))

        self.net.load_state_dict(best_model_wts)

    def _reconstruct(self, observation):
        self.net.eval()
        fbp = self.fbp_op(observation)
        tensor_type = (torch.cuda.FloatTensor if self.use_cuda else
                       torch.FloatTensor)
        fbp_tensor = (torch.from_numpy(np.asarray(fbp)[None, None])
                      .type(tensor_type))
        reco_tensor = self.net(fbp_tensor)
        reconstruction = reco_tensor.cpu().detach().numpy()[0, 0]
        return self.reco_space.element(reconstruction)

    def save_params(self, filename):
        """
        Save U-Net state to file using :meth:`torch.save`.

        Parameters
        ----------
        filename : str
            Filename (ending ``.pt``).
        """
        torch.save(self.net.state_dict(), filename)

    def load_params(self, filename):
        """
        Load U-Net state from file using :meth:`torch.load`.

        Parameters
        ----------
        filename : str
            Filename (ending ``.pt``).
        """
        map_location = 'cuda:0' if self.use_cuda else 'cpu'
        self.net.load_state_dict(torch.load(filename,
                                            map_location=map_location))


def get_skip_model(scales=5, skip=4):
    assert(1 <= scales <= 6)
    channels = (16, 32, 64, 64, 128, 128)
    skip_channels = [skip] * (scales)
    return Skip(in_ch=1, out_ch=1, channels=channels[:scales],
                skip_channels=skip_channels)


class Skip(nn.Module):
    def __init__(self, in_ch, out_ch, channels, skip_channels):
        super(Skip, self).__init__()
        assert(len(channels) == len(skip_channels))
        self.scales = len(channels)

        self.down = nn.ModuleList()
        self.up = nn.ModuleList()
        self.inc = InBlock(in_ch, channels[0])
        for i in range(1, self.scales):
            self.down.append(DownBlock(in_ch=channels[i-1],
                                       out_ch=channels[i]))
        for i in range(1, self.scales):
            self.up.append(UpBlock(in_ch=channels[-i], out_ch=channels[-i-1],
                                   skip_ch=skip_channels[-i]))

        self.outc = OutBlock(in_ch=channels[0], out_ch=out_ch)

    def forward(self, x0):
        xs = [self.inc(x0), ]
        for i in range(self.scales-1):
            xs.append(self.down[i](xs[-1]))
        x = xs[-1]
        for i in range(self.scales-1):
            x = self.up[i](x, xs[-2-i])

        return torch.sigmoid(self.outc(x))


class DownBlock(nn.Module):
    def __init__(self, in_ch, out_ch, kernel_size=3):
        super(DownBlock, self).__init__()
        to_pad = int((kernel_size - 1) / 2)
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size, stride=2, padding=to_pad),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(out_ch, out_ch, kernel_size, stride=1, padding=to_pad),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(0.2, inplace=True))

    def forward(self, x):
        x = self.conv(x)
        return x


class InBlock(nn.Module):
    def __init__(self, in_ch, out_ch, kernel_size=3):
        super(InBlock, self).__init__()
        to_pad = int((kernel_size - 1) / 2)
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size, stride=1, padding=to_pad),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(0.2, inplace=True))

    def forward(self, x):
        x = self.conv(x)
        return x


class UpBlock(nn.Module):
    def __init__(self, in_ch, out_ch, skip_ch=4, kernel_size=3):
        super(UpBlock, self).__init__()
        to_pad = int((kernel_size - 1) / 2)
        self.skip = skip_ch > 0
        if skip_ch == 0:
            skip_ch = 1
        self.conv = nn.Sequential(
            nn.BatchNorm2d(in_ch + skip_ch),
            nn.Conv2d(in_ch + skip_ch, out_ch, kernel_size, stride=1,
                      padding=to_pad),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Conv2d(out_ch, out_ch, kernel_size, stride=1, padding=to_pad),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(0.2, inplace=True))

        self.skip_conv = nn.Sequential(
            nn.Conv2d(out_ch, skip_ch, kernel_size=1, stride=1),
            nn.BatchNorm2d(skip_ch),
            nn.LeakyReLU(0.2, inplace=True))

        self.up = nn.Upsample(scale_factor=2, mode='bilinear',
                              align_corners=True)
        self.concat = Concat()

    def forward(self, x1, x2):
        x1 = self.up(x1)
        x2 = self.skip_conv(x2)
        if not self.skip:
            x2 = x2 * 0
        x = self.concat(x1, x2)
        x = self.conv(x)
        return x


class Concat(nn.Module):
    def __init__(self):
        super(Concat, self).__init__()

    def forward(self, *inputs):
        inputs_shapes2 = [x.shape[2] for x in inputs]
        inputs_shapes3 = [x.shape[3] for x in inputs]

        if (np.all(np.array(inputs_shapes2) == min(inputs_shapes2)) and
                np.all(np.array(inputs_shapes3) == min(inputs_shapes3))):
            inputs_ = inputs
        else:
            target_shape2 = min(inputs_shapes2)
            target_shape3 = min(inputs_shapes3)

            inputs_ = []
            for inp in inputs:
                diff2 = (inp.size(2) - target_shape2) // 2
                diff3 = (inp.size(3) - target_shape3) // 2
                inputs_.append(inp[:, :, diff2: diff2 + target_shape2,
                                   diff3:diff3 + target_shape3])
        return torch.cat(inputs_, dim=1)


class OutBlock(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(OutBlock, self).__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1)

    def forward(self, x):
        x = self.conv(x)
        return x

    def __len__(self):
        return len(self._modules)



from dival.util.odl_utility import uniform_discr_element
from dival.data import DataPairs
from dival.evaluation import TaskTable
from dival.measure import L2
from dival import Reconstructor
import numpy as np

np.random.seed(1)

ground_truth = uniform_discr_element([0, 1, 2, 3, 4, 5, 6])
observation = ground_truth + 1
observation += np.random.normal(size=observation.shape)
test_data = DataPairs(observation, ground_truth, name='x + 1 + normal')
eval_tt = TaskTable()


class MinusOneReconstructor(Reconstructor):
    def reconstruct(self, observation):
        return observation - 1


reconstructor = MinusOneReconstructor(name='y-1')
eval_tt.append(reconstructor=reconstructor, test_data=test_data,
               measures=[L2])
results = eval_tt.run()
results.plot_reconstruction(0)
print(results)



# -*- coding: utf-8 -*-
"""Provides utilities related to ODL."""
from dival.util.odl_noise_random_state import (white_noise, uniform_noise,
                                               poisson_noise,
                                               salt_pepper_noise)
from odl.discr.lp_discr import uniform_discr
from odl.operator.operator import Operator
import numpy as np


def uniform_discr_element(inp, space=None):
    """Generate an element of a ODL space from an array-like.

    Parameters
    ----------
    inp : array-like
        The input data from which the element is generated.
    space : :class:`odl.discr.DiscretizedSpace`, optional
        The space which the element will belong to. If not given, a uniform
        discretization space with cell size 1 centered around the origin is
        generated.
    """
    inp = np.asarray(inp)
    if space is None:
        space = uniform_discr(-np.array(inp.shape)/2, np.array(inp.shape)/2,
                              inp.shape)
    element = space.element(inp)
    return element


def apply_noise(x, noise_type, noise_kwargs=None, seed=None,
                random_state=None):
    """Apply noise to an odl element.

    Calls noise functions from :mod:`odl.phantom.noise` or their equivalents
    from :mod:`dival.util.odl_noise_random_state`.

    Parameters
    ----------
    x : odl element
        The element to which the noise is applied (in-place).
    noise_type : {``'white'``, ``'uniform'``, ``'poisson'``, ``'salt_pepper'``}
        Type of noise.
    noise_kwargs : dict, optional
        Keyword arguments to be passed to the noise function, e.g. ``'stddev'``
        for ``'white'`` noise.
        The arguments are:

            * for ``noise_type='white'``:
                * ``'stddev'``: float, optional
                    Standard deviation of each component of the normal
                    distribution. Default is 1.
                * ``'relative_stddev'``: bool, optional
                    Whether to multiply ``'stddev'`` with ``mean(abs(x))``.
                    Default is ``False``.
            * for ``noise_type='poisson'``:
                * ``'scaling_factor'``: float, optional
                    If specified, the intensity is multiplied and the samples
                    from the poisson distribution are divided by this factor:
                    ``poisson(x * scaling_factor) / scaling_factor``.
                    Default is `None`.
    seed : int, optional
        Random seed passed to the noise function.
    random_state : :class:`np.random.RandomState`, optional
        Random state passed to the noise function.
    """
    n_kwargs = noise_kwargs.copy()
    n_kwargs['seed'] = seed
    n_kwargs['random_state'] = random_state
    if noise_type == 'white':
        relative_stddev = n_kwargs.pop('relative_stddev', False)
        stddev = n_kwargs.pop('stddev', 1.)
        if relative_stddev:
            mean_abs = np.mean(np.abs(x))
            stddev *= mean_abs
        noise = white_noise(x.space, stddev=stddev, **n_kwargs)
        x += noise
    elif noise_type == 'uniform':
        noise = uniform_noise(x.space, **n_kwargs)
        x += noise
    elif noise_type == 'poisson':
        scaling_factor = n_kwargs.pop('scaling_factor', None)
        if scaling_factor:
            x.assign(poisson_noise(x * scaling_factor, **n_kwargs) /
                     scaling_factor)
        else:
            x.assign(poisson_noise(x, **n_kwargs))
    elif noise_type == 'salt_pepper':
        noise = salt_pepper_noise(x.domain, **n_kwargs)
        x += noise
    else:
        raise ValueError("unknown noise type '{}'".format(noise_type))


class NoiseOperator(Operator):
    """Operator applying noise.

    Wraps :func:`apply_noise`, which calls noise functions from
    :mod:`odl.phantom.noise` or their equivalents from
    :mod:`dival.util.odl_noise_random_state`.
    """
    def __init__(self, domain, noise_type, noise_kwargs=None, seed=None,
                 random_state=None):
        """
        Parameters
        ----------
        space : odl space
            Domain and range.
        noise_type : {``'white'``, ``'uniform'``, ``'poisson'``,\
                      ``'salt_pepper'``}
            Type of noise.
        noise_kwargs : dict, optional
            Keyword arguments to be passed to the noise function, cf. docs for
            :func:`apply_noise`.
        seed : int, optional
            Random seed passed to the noise function.
        random_state : `np.random.RandomState`, optional
            Random state passed to the noise function.
        """
        super().__init__(domain, domain)
        self.noise_type = noise_type or 'white'
        self.noise_kwargs = noise_kwargs or {}
        self.seed = seed
        self.random_state = random_state

    def _call(self, x, out):
        if out is not x:
            out.assign(x)
        apply_noise(out, self.noise_type, noise_kwargs=self.noise_kwargs,
                    seed=self.seed, random_state=self.random_state)



# -*- coding: utf-8 -*-
"""Provides reconstructors performing regression."""
import os
import json
import numpy as np
from sklearn.linear_model import Ridge
from dival.util.odl_utility import uniform_discr_element
from dival import LearnedReconstructor


class LinRegReconstructor(LearnedReconstructor):
    HYPER_PARAMS = {
        'l2_regularization':
            {'default': 0.,
             'range': [0., np.inf],
             'retrain': True}
    }

    """Reconstructor learning and applying linear regression.

    Assumes the inverse operator is linear, i.e. ``x = A_inv * y``.
    Learns the entries of ``A_inv`` by l2-regularized linear regression:
    ``A_inv = 1/2N * sum_i ||x_i - A_inv * y_i||^2 + alpha/2 * ||A_inv||_F^2``,
    where (y_i, x_i) with i=1,...,N are pairs of observations and the
    corresponding ground truth.

    Attributes
    ----------
    weights : :class:`np.ndarray`
        The weight matrix.
    """
    def __init__(self, hyper_params=None, **kwargs):
        """
        Parameters
        ----------
        hyper_params : dict, optional
            A dict with no items or an item ``'l2_regularization': float``.
            Cf. :meth:`Reconstructor.init`.
        """
        super().__init__(hyper_params=hyper_params, **kwargs)
        self.weights = None

    def _reconstruct(self, observation):
        reconstruction = np.dot(self.weights, observation)
        return uniform_discr_element(reconstruction, self.reco_space)

    def train(self, dataset):
        observation_shape, reco_shape = dataset.get_shape()
        if (self.observation_space is not None
                and self.observation_space.shape != observation_shape):
            raise ValueError('Observation shape of dataset not matching '
                             '`self.observation_space.shape`')
        if (self.observation_space is not None
                and self.observation_space.shape != observation_shape):
            raise ValueError('Observation shape of dataset not matching '
                             '`self.observation_space.shape`')
        ridge = Ridge(self.hyper_params['l2_regularization'])
        train_len = dataset.get_train_len()
        n_features = dataset.shape[0][0]
        n_targets = dataset.shape[1][0]
        x = np.empty((train_len, n_features))
        y = np.empty((train_len, n_targets))
        for i, (x_i, y_i) in enumerate(dataset.get_train_generator()):
            x[i] = x_i
            y[i] = y_i
        ridge.fit(x, y)
        self.weights = ridge.coef_

    def save_params(self, path):
        """
        Save :attr:`weights` to file.

        Parameters
        ----------
        path : str
            Filename (ending '.npy').
        """
        if not os.path.isdir(path):
            os.mkdir(path)
        np.save(os.path.join(path, 'weights.npy'), self.weights)
        with open(os.path.join(path, 'hyper_params.json'), 'w') as file:
            json.dump(self.hyper_params, file, indent=True)

    def load_params(self, path):
        """
        Load :attr:`weights` from file.

        Parameters
        ----------
        path : str
            Filename (ending '.npy').
        """
        self.weights = np.load(os.path.join(path, 'weights.npy'))
        with open(os.path.join(path, 'hyper_params.json'), 'r') as file:
            self.hyper_params.update(json.load(file))



# -*- coding: utf-8 -*-
"""Provides the dataset base classes.
"""
from itertools import islice
from math import ceil
import numpy as np
from dival.data import DataPairs
from dival.util.odl_utility import NoiseOperator


class Dataset():
    """Dataset base class.

    Attributes
    ----------
    space : [tuple of ] :class:`odl.space.base_tensors.TensorSpace` or `None`
        The spaces of the elements of samples as a tuple.
        If only one element per sample is provided, this attribute is the space
        of the element (i.e., no tuple).
        It is strongly recommended to set this attribute in subclasses, as some
        functionality may depend on it.
    shape : [tuple of ] tuple of int, optional
        The shapes of the elements of samples as a tuple of tuple of int.
        If only one element per sample is provided, this attribute is the shape
        of the element (i.e., not a tuple of tuple of int, but a tuple of int).
    train_len : int, optional
        Number of training samples.
    validation_len : int, optional
        Number of validation samples.
    test_len : int, optional
        Number of test samples.
    random_access : bool, optional
        Whether the dataset supports random access via ``self.get_sample`` and
        ``self.get_samples``.
        Setting this attribute is the preferred way for subclasses to indicate
        whether they support random access.
    num_elements_per_sample : int, optional
        Number of elements per sample.
        E.g. 1 for a ground truth dataset or 2 for a dataset of pairs of
        observation and ground truth.
    standard_dataset_name : str, optional
        Datasets returned by `get_standard_dataset` have this attribute giving
        its name.
    """
    def __init__(self, space=None):
        """
        Parameters
        ----------
        space : [tuple of ] :class:`odl.space.base_tensors.TensorSpace`,\
                optional
            The spaces of the elements of samples as a tuple.
            If only one element per sample is provided, this attribute is the
            space of the element (i.e., no tuple).
            It is strongly recommended to set `space` in subclasses, as some
            functionality may depend on it.
        """
        self.space = space

    def generator(self, part='train'):
        """Yield data.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            Whether to yield train, validation or test data.
            Default is ``'train'``.

        Yields
        ------
        data : odl element or tuple of odl elements
            Sample of the dataset.
        """
        if self.supports_random_access():
            for i in range(self.get_len(part)):
                yield self.get_sample(i, part=part)
        else:
            raise NotImplementedError

    def get_train_generator(self):
        return self.generator(part='train')

    def get_validation_generator(self):
        return self.generator(part='validation')

    def get_test_generator(self):
        return self.generator(part='test')

    def get_len(self, part='train'):
        """Return the number of elements the generator will yield.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            Whether to return the number of train, validation or test elements.
            Default is ``'train'``.
        """
        if part == 'train':
            return self.get_train_len()
        elif part == 'validation':
            return self.get_validation_len()
        elif part == 'test':
            return self.get_test_len()
        else:
            raise ValueError("dataset part must be 'train', "
                             "'validation' or 'test', not '{}'".format(part))

    def get_train_len(self):
        """Return the number of samples the train generator will yield."""
        try:
            return self.train_len
        except AttributeError:
            raise NotImplementedError

    def get_validation_len(self):
        """Return the number of samples the validation generator will yield.
        """
        try:
            return self.validation_len
        except AttributeError:
            raise NotImplementedError

    def get_test_len(self):
        """Return the number of samples the test generator will yield."""
        try:
            return self.test_len
        except AttributeError:
            raise NotImplementedError

    def get_shape(self):
        """Return the shape of each element."""
        try:
            return self.shape
        except AttributeError:
            raise NotImplementedError

    def get_num_elements_per_sample(self):
        """Return number of elements per sample.

        If :attr:`num_elements_per_sample` is not set, an attempt to
        infer it by calling :meth:`get_shape` is made.
        """
        try:
            return self.num_elements_per_sample
        except AttributeError:
            shape = self.get_shape()  # may raise NotImplementedError
            return len(shape) if isinstance(shape, tuple) else 1

    def get_data_pairs(self, part='train', n=None):
        """
        Return first samples from data part as :class:`.DataPairs` object.

        Only supports datasets with two elements per sample.``

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        n : int, optional
            Number of pairs (from beginning). If `None`, all available data
            is used (the default).
        """
        if self.get_num_elements_per_sample() != 2:
            raise ValueError('`get_data_pairs` only supports datasets with'
                             '2 elements per sample, this dataset has {:d}'
                             .format(self.get_num_elements_per_sample()))
        gen = self.generator(part=part)
        observations, ground_truth = [], []
        for obs, gt in islice(gen, n):
            observations.append(obs)
            ground_truth.append(gt)
        name = '{} part{}'.format(part,
                                  ' 0:{:d}'.format(n) if n is not None else '')
        data_pairs = DataPairs(observations, ground_truth, name=name)
        return data_pairs

    def create_torch_dataset(self, part='train', reshape=None):
        """
        Create a torch dataset wrapper for one part of this dataset.

        If :meth:`supports_random_access` returns ``False``, samples are
        fetched from :meth:`generator`. The index passed to
        :meth:`~torch.utils.data.dataset.Dataset.__getitem__` of the
        returned dataset will be ignored, and parallel data loading (with
        multiple workers) is not applicable.

        If :meth:`supports_random_access` returns `True`, samples are looked
        up using :meth:`get_sample`. For datasets that support parallel calls
        to :meth:`get_sample`, the returned torch dataset can be used by
        multiple workers.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        reshape : tuple of (tuple or `None`), optional
            Shapes to which the elements of each sample will be reshaped.
            If `None` is passed for an element, no reshape is applied.
        """
        from torch.utils.data import Dataset as TorchDataset
        import torch

        if self.supports_random_access():
            class RandomAccessTorchDataset(TorchDataset):
                def __init__(self, dataset, part, reshape=None):
                    self.dataset = dataset
                    self.part = part
                    self.reshape = reshape or (
                        (None,) * self.dataset.get_num_elements_per_sample())

                def __len__(self):
                    return self.dataset.get_len(self.part)

                def __getitem__(self, idx):
                    arrays = self.dataset.get_sample(idx, part=self.part)
                    mult_elem = isinstance(arrays, tuple)
                    if not mult_elem:
                        arrays = (arrays,)
                    tensors = []
                    for arr, s in zip(arrays, self.reshape):
                        t = torch.from_numpy(np.asarray(arr))
                        if s is not None:
                            t = t.view(*s)
                        tensors.append(t)
                    return tuple(tensors) if mult_elem else tensors[0]

            dataset = RandomAccessTorchDataset(self, part, reshape=reshape)
        else:
            class GeneratorTorchDataset(TorchDataset):
                def __init__(self, dataset, part, reshape=None):
                    self.generator = dataset.generator(part)
                    self.length = dataset.get_len(part)
                    self.reshape = reshape or (
                        (None,) * self.dataset.get_num_elements_per_sample())

                def __len__(self):
                    return self.length

                def __getitem__(self, idx):
                    arrays = next(self.generator)
                    mult_elem = isinstance(arrays, tuple)
                    if not mult_elem:
                        arrays = (arrays,)
                    tensors = []
                    for arr, s in zip(arrays, self.reshape):
                        t = torch.from_numpy(np.asarray(arr))
                        if s is not None:
                            t = t.view(*s)
                        tensors.append(t)
                    return tuple(tensors) if mult_elem else tensors[0]

            dataset = GeneratorTorchDataset(self, part, reshape=reshape)

        return dataset

    def create_keras_generator(self, part='train', batch_size=1, shuffle=True,
                               reshape=None):
        """
        Create a keras data generator wrapper for one part of this dataset.

        If :meth:`supports_random_access` returns ``False``, a generator
        wrapping :meth:`generator` is returned. In this case no shuffling is
        performed regardless of the passed `shuffle` parameter. Also, parallel
        data loading (with multiple workers) is not applicable.

        If :meth:`supports_random_access` returns `True`, a
        :class:`tf.keras.utils.Sequence` is returned, which is implemented
        using :meth:`get_sample`. For datasets that support parallel calls to
        :meth:`get_sample`, the returned data generator (sequence) can be used
        by multiple workers.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        batch_size : int, optional
            Batch size. Default is 1.
        shuffle : bool, optional
            Whether to shuffle samples each epoch.
            This option has no effect if :meth:`supports_random_access` returns
            ``False``, since in that case samples are fetched directly from
            :meth:`generator`.
            The default is `True`.
        reshape : tuple of (tuple or `None`), optional
            Shapes to which the elements of each sample will be reshaped.
            If `None` is passed for an element, no reshape is applied.
        """
        from tensorflow.keras.utils import Sequence

        if self.supports_random_access():
            class KerasGenerator(Sequence):
                def __init__(self, dataset, part, batch_size, shuffle,
                             reshape=None):
                    self.dataset = dataset
                    self.part = part
                    self.batch_size = batch_size
                    self.shuffle = shuffle
                    self.reshape = reshape or (
                        (None,) * self.dataset.get_num_elements_per_sample())
                    self.data_shape = self.dataset.get_shape()
                    self.on_epoch_end()

                def __len__(self):
                    return ceil(self.dataset.get_len(self.part) /
                                self.batch_size)

                def __getitem__(self, idx):
                    indexes = self.indexes[idx*self.batch_size:
                                           (idx+1)*self.batch_size]
                    # for last batch, indexes has len <= batch_size
                    n_elem = self.dataset.get_num_elements_per_sample()
                    arrays = []
                    for i in range(n_elem):
                        array = np.empty(
                            (len(indexes),) + self.data_shape[i],
                            dtype=self.dataset.space[i].dtype)
                        arrays.append(array)
                    for j, ind in enumerate(indexes):
                        out = tuple([array[j] for array in arrays])
                        self.dataset.get_sample(ind, part=self.part, out=out)
                    for i in range(n_elem):
                        if self.reshape[i] is not None:
                            arrays[i] = arrays[i].reshape(
                                (len(indexes),) + self.reshape[i])
                    return tuple(arrays) if n_elem > 1 else arrays[0]

                def on_epoch_end(self):
                    self.indexes = np.arange(self.dataset.get_len(self.part))
                    if self.shuffle:
                        np.random.shuffle(self.indexes)

            generator = KerasGenerator(self, part, batch_size=batch_size,
                                       shuffle=shuffle, reshape=reshape)

        else:
            def keras_generator(dataset, part, batch_size, reshape=None):
                generator = dataset.generator(part)
                n_elem = dataset.get_num_elements_per_sample()
                num_steps_per_epoch = ceil(dataset.get_len(part) / batch_size)
                if reshape is None:
                    reshape = (None,) * n_elem
                data_shape = dataset.get_shape()
                while True:
                    for k in range(num_steps_per_epoch):
                        batch_size_ = (batch_size if k < num_steps_per_epoch-1
                                       else dataset.get_len(part) % batch_size)
                        arrays = []
                        for i in range(n_elem):
                            array = np.empty(
                                (batch_size_,) + data_shape[i],
                                dtype=dataset.space[i].dtype)
                            arrays.append(array)
                        for j in range(batch_size_):
                            sample = next(generator)
                            if n_elem == 1:
                                sample = (sample,)
                            for i, array in enumerate(arrays):
                                array[j, :] = sample[i]
                        for i in range(n_elem):
                            if reshape[i] is not None:
                                arrays[i] = arrays[i].reshape(
                                    (batch_size_,) + reshape[i])
                        yield tuple(arrays) if n_elem > 1 else arrays[0]

            generator = keras_generator(self, part, batch_size=batch_size,
                                        reshape=reshape)

        return generator

    def get_sample(self, index, part='train', out=None):
        """Get single sample by index.

        Parameters
        ----------
        index : int
            Index of the sample.
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        out : array-like or tuple of (array-like or bool) or `None`
            Array(s) (or e.g. odl element(s)) to which the sample is written.
            A tuple should be passed, if the dataset returns two or more arrays
            per sample (i.e. pairs, ...).
            If a tuple element is a bool, it has the following meaning:

                ``True``
                    Create a new array and return it.
                ``False``
                    Do not return this array, i.e. `None` is returned.

        Returns
        -------
        sample : [tuple of ] (array-like or `None`)
            E.g. for a pair dataset: ``(array, None)`` if
            ``out=(True, False)``.
        """
        raise NotImplementedError

    def get_samples(self, key, part='train', out=None):
        """Get samples by slice or range.

        Parameters
        ----------
        key : slice or range
            Indexes of the samples.
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        out : array-like or tuple of (array-like or bool) or `None`
            Array(s) (or e.g. odl element(s)) to which the sample is written.
            The first dimension must match the number of samples requested.
            A tuple should be passed, if the dataset returns two or more arrays
            per sample (i.e. pairs, ...).
            If a tuple element is a bool, it has the following meaning:

                ``True``
                    Create a new array and return it.
                ``False``
                    Do not return this array, i.e. `None` is returned.

        Returns
        -------
        samples : [tuple of ] (array-like or `None`)
            If the dataset has multiple arrays per sample, a tuple holding
            arrays is returned.
            E.g. for a pair dataset: ``(array, None)`` if
            ``out=(True, False)``.
            The samples are stacked in the first
            (additional) dimension of each array.
        """
        raise NotImplementedError

    def supports_random_access(self):
        """Whether random access seems to be supported.

        If the object has the attribute `self.random_access`, its value is
        returned (this is the preferred way for subclasses to indicate whether
        they support random access). Otherwise, a simple duck-type check is
        performed which tries to get the first sample by random access.

        Returns
        -------
        supports : bool
            ``True`` if the dataset supports random access, otherwise
            ``False``.
        """
        try:
            return self.random_access
        except AttributeError:
            try:
                self.get_sample(0)
                self.get_samples(slice(1))
            except NotImplementedError:
                return False
            return True


class ObservationGroundTruthPairDataset(Dataset):
    """
    Dataset of pairs generated from a ground truth generator by applying a
    forward operator and noise.
    """
    def __init__(self, ground_truth_gen, forward_op, post_processor=None,
                 train_len=None, validation_len=None, test_len=None,
                 domain=None, noise_type=None, noise_kwargs=None,
                 noise_seeds=None):
        """
        Parameters
        ----------
        ground_truth_gen : generator function
            Function returning a generator providing ground truth.
            Must accept a `part` parameter like :meth:`Dataset.generator`.
        forward_op : odl operator
            Forward operator to apply on the ground truth.
        post_processor : odl operator, optional
            Post-processor to apply on the result of the forward operator.
        train_len : int, optional
            Number of training samples.
        validation_len : int, optional
            Number of validation samples.
        test_len : int, optional
            Number of test samples.
        domain : odl space, optional
            Ground truth domain.
            If not specified, it is inferred from `forward_op`.
        noise_type : str, optional
            Noise type. See :class:`~dival.util.odl_utility.NoiseOperator` for
            the list of supported noise types.
        noise_kwargs : dict, optional
            Keyword arguments passed to
            :class:`~dival.util.odl_utility.NoiseOperator`.
        noise_seeds : dict of int, optional
            Seeds to use for random noise generation.
            The part (``'train'``, ...) is the key to the dict.
            If a key is omitted or a value is `None`, no fixed seed is used
            for that part. By default, no fixed seeds are used.
        """
        self.ground_truth_gen = ground_truth_gen
        self.forward_op = forward_op
        self.post_processor = post_processor
        if train_len is not None:
            self.train_len = train_len
        if validation_len is not None:
            self.validation_len = validation_len
        if test_len is not None:
            self.test_len = test_len
        if domain is None:
            domain = self.forward_op.domain
        self.noise_type = noise_type
        self.noise_kwargs = noise_kwargs
        self.noise_seeds = noise_seeds or {}
        range_ = (self.post_processor.range if self.post_processor is not None
                  else self.forward_op.range)
        super().__init__(space=(range_, domain))
        self.shape = (self.space[0].shape, self.space[1].shape)
        self.num_elements_per_sample = 2

    def generator(self, part='train'):
        gt_gen_instance = self.ground_truth_gen(part=part)
        if self.noise_type is not None:
            random_state = np.random.RandomState(self.noise_seeds.get(part))
            noise_op = NoiseOperator(self.forward_op.range, self.noise_type,
                                     noise_kwargs=self.noise_kwargs,
                                     random_state=random_state)
            full_op = noise_op * self.forward_op
        else:
            full_op = self.forward_op
        if self.post_processor is not None:
            full_op = self.post_processor * full_op
        for ground_truth in gt_gen_instance:
            yield (full_op(ground_truth), ground_truth)


class GroundTruthDataset(Dataset):
    """
    Ground truth dataset base class.
    """
    def __init__(self, space=None):
        """
        Parameters
        ----------
        space : :class:`odl.space.base_tensors.TensorSpace`, optional
            The space of the samples.
            It is strongly recommended to set `space` in subclasses, as some
            functionality may depend on it.
        """
        self.num_elements_per_sample = 1
        super().__init__(space=space)

    def create_pair_dataset(self, forward_op, post_processor=None,
                            noise_type=None, noise_kwargs=None,
                            noise_seeds=None):
        """
        The parameters are a subset of those of
        :meth:`ObservationGroundTruthPairDataset.__init__`.
        """
        try:
            train_len = self.get_train_len()
        except NotImplementedError:
            train_len = None
        try:
            validation_len = self.get_validation_len()
        except NotImplementedError:
            validation_len = None
        try:
            test_len = self.get_test_len()
        except NotImplementedError:
            test_len = None
        dataset = ObservationGroundTruthPairDataset(
            self.generator, forward_op, post_processor=post_processor,
            train_len=train_len, validation_len=validation_len,
            test_len=test_len, noise_type=noise_type,
            noise_kwargs=noise_kwargs, noise_seeds=noise_seeds)
        return dataset



# -*- coding: utf-8 -*-
"""
Configuration of the library.

The module variable :obj:`CONFIG` holds the current configuration.

The library configuration is stored in the file ``~/.dival/config.json``.
If the config file does not exist when the library is imported, it is created
using default values.
Changes made in ``config.json`` are loaded when the library is (re-)imported.
"""
import os
import json

CONFIG = {
 'lodopab_dataset': {
  'data_path': os.path.normpath(
      os.path.expanduser('~/.dival/datasets/lodopab'))
 }
}
"""
Global configuration dict.

Holds the current configuration of the library. On ``import dival``, the
configuration is loaded from ~/.dival/config.json.
"""

# configuration loaded from config file, overrides hardcoded configuration
CONFIG_FILENAME = os.path.normpath(os.path.expanduser('~/.dival/config.json'))

# automatically write config file if not existing (e.g., on first import)
if not os.path.isfile(CONFIG_FILENAME):
    os.makedirs(os.path.dirname(CONFIG_FILENAME), exist_ok=True)
    with open(CONFIG_FILENAME, 'w') as config_fp:
        json.dump(CONFIG, config_fp, indent=1)
with open(CONFIG_FILENAME, 'r') as config_fp:
    CONFIG.update(json.load(config_fp))


def get_config(key_path='/'):
    """
    Return (sub-)configuration stored in config file.
    Note that values may differ from the current ``CONFIG`` variable if it was
    manipulated directly.

    Parameters
    ----------
    key_path : str, optional
        ``'/'``-separated path to sub-configuration. Default is ``'/'``, which
        returns the full configuration dict.

    Returns
    -------
    sub_config
        (sub-)configuration, either a dict or a value
    """
    keys = [k for k in key_path.split('/') if k != '']
    with open(CONFIG_FILENAME, 'r') as config_fp:
        config = json.load(config_fp)
    sub_config = config
    for k in keys:
        sub_config = sub_config[k]
    return sub_config


def set_config(key_path, value, verbose=True):
    """
    Updates (sub-)configuration both in ``CONFIG`` variable and in config file.

    Parameters
    ----------
    key_path : str, optional
        ``'/'``-separated path to sub-configuration. Pass ``'/'`` to replace
        the full configuration dict.
    value : object
        (sub-)configuration value. Either a dict, which is copied, or a value.
    """
    global CONFIG
    c = get_config()
    keys = [k for k in key_path.split('/') if k != '']
    if isinstance(value, dict):
        value = value.copy()
    if len(keys) == 0:
        CONFIG = value
        c = value
    else:
        sub_config = CONFIG
        sub_c = c
        for k in keys[:-1]:
            sub_config = sub_config[k]
            sub_c = sub_c[k]
        sub_config[keys[-1]] = value
        sub_c[keys[-1]] = value
    with open(CONFIG_FILENAME, 'w') as config_fp:
        json.dump(c, config_fp, indent=1)
    if verbose:
        print("updated configuration in '{}':".format(CONFIG_FILENAME))
        print("'{}' = {}".format(key_path, value))



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from dival import __version__

setup(name='dival',
      version=__version__,
      description='Deep Inversion Validation Library',
      url='https://github.com/jleuschn/dival',
      author='Johannes Leuschner',
      author_email='jleuschn@uni-bremen.de',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy>=1.10',
          'pandas',
          'odl',
          'scikit-image',
          'scikit-learn',
          'hyperopt',
          'pydicom',
          'tqdm',
          'matplotlib'
      ],
      zip_safe=False)



# -*- coding: utf-8 -*-
"""
Provides physical and numerical constants.
"""

MU_WATER = 20  # M^-1
"""
Linear attenuation of water in SI unit ``m^-1``.
Source:
`nist.gov <https://www.nist.gov/pml/x-ray-mass-attenuation-coefficients>`_.
"""

MU_AIR = 0.02  # M^-1
"""
Linear attenuation of air in SI unit ``m^-1``.
Source:
`nist.gov <https://www.nist.gov/pml/x-ray-mass-attenuation-coefficients>`_.
"""

MU_MAX = 3071 * (MU_WATER - MU_AIR) / 1000 + MU_WATER
"""
Maximum linear attenuation representable with standard 12 bit HU value format.
The value is in SI unit ``m^-1``.
"""



# -*- coding: utf-8 -*-
"""Provides `LoDoPaBDataset`."""
import os
from warnings import warn
from math import ceil
import numpy as np
import h5py
from zipfile import ZipFile
from tqdm import tqdm
from odl.discr.lp_discr import uniform_discr
import odl.tomo
from dival.datasets.dataset import Dataset
from dival.config import CONFIG, set_config
from dival.util.constants import MU_MAX
from dival.util.zenodo_download import download_zenodo_record, input_yes_no


DATA_PATH = CONFIG['lodopab_dataset']['data_path']
NUM_SAMPLES_PER_FILE = 128
PHOTONS_PER_PIXEL = 4096
ORIG_MIN_PHOTON_COUNT = 0.1
MIN_PT = [-0.13, -0.13]
MAX_PT = [0.13, 0.13]


def download_lodopab():
    global DATA_PATH
    print('Before downloading, please make sure to have enough free disk '
          'space (~150GB). After unpacking, 114.7GB will be used.')
    print("path to store LoDoPaB-CT dataset (default '{}'):".format(DATA_PATH))
    inp = input()
    if inp:
        DATA_PATH = inp
        set_config('lodopab_dataset/data_path', DATA_PATH)
    os.makedirs(DATA_PATH, exist_ok=True)
    ZENODO_RECORD_ID = '3384092'
    success = download_zenodo_record(ZENODO_RECORD_ID, DATA_PATH)
    print('download of LoDoPaB-CT dataset {}'.format('successful' if success
                                                     else 'failed'))
    if not success:
        return False
    file_list = ['observation_train.zip',      'ground_truth_train.zip',
                 'observation_validation.zip', 'ground_truth_validation.zip',
                 'observation_test.zip',       'ground_truth_test.zip']
    print('unzipping zip files, this can take several minutes', flush=True)
    for file in tqdm(file_list, desc='unzip'):
        filename = os.path.join(DATA_PATH, file)
        with ZipFile(filename, 'r') as f:
            f.extractall(DATA_PATH)
        os.remove(filename)
    return True


class LoDoPaBDataset(Dataset):
    """
    A simulated low dose CT dataset based on real reconstructions from the
    `LIDC-IDRI
    <https://wiki.cancerimagingarchive.net/display/Public/LIDC-IDRI>`_ dataset.

    The dataset contains 42895 pairs of images and projection data.
    For simulation, a ray transform with parallel beam geometry using 1000
    angles and 513 detector pixels is used. Poisson noise corresponding to 4096
    incident photons per pixel before attenuation is applied to the projection
    data. The images have a size of 362x362 px.

    A ray transform that corresponds to the noiseless forward operator is
    available via the `get_ray_trafo` method of this dataset.

    Attributes
    ----------
    space
        ``(space[0], space[1])``, where
            ``space[0]``
                ``odl.uniform_discr([0., -0.1838], [3.1416, 0.1838],
                (1000, 513), dtype='float32')``
            ``space[1]``
                ``odl.uniform_discr(min_pt, max_pt, (362, 362),
                dtype='float32'))``, with `min_pt` and `max_pt` parameters
                passed to :meth:`__init__`
    shape
        ``(362, 362)``
    train_len
        ``35820``
    validation_len
       ``3522``
    test_len
        ``3553``
    random_access
        ``True``
    num_elements_per_sample
        ``2``
    """
    def __init__(self, min_pt=None, max_pt=None, observation_model='post-log',
                 min_photon_count=None):
        """
        Parameters
        ----------
        min_pt : [float, float], optional
            Minimum values of the lp space. Default: ``[-0.13, -0.13]``.
        max_pt : [float, float], optional
            Maximum values of the lp space. Default: ``[0.13, 0.13]``.
        observation_model : {'post-log', 'pre-log'}, optional
            The observation model to use.
            The default is ``'post-log'``.

            ``'post-log'``
                Observations are linearly related to the normalized ground
                truth via the ray transform, ``obs = ray_trafo(gt) + noise``.
                Note that the scaling of the observations matches the
                normalized ground truth, i.e., they are divided by the linear
                attenuation of 3071 HU.
            ``'pre-log'``
                Observations are non-linearly related to the ground truth, as
                given by the Beer-Lambert law.
                The model is
                ``obs = exp(-ray_trafo(gt * MU(3071 HU))) + noise``,
                where `MU(3071 HU)` is the factor, by which the ground truth
                was normalized.
        min_photon_count : float, optional
            Replacement value for a simulated photon count of zero.
            If ``observation_model == 'post-log'``, a value greater than zero
            is required in order to avoid undefined values. The default is 0.1,
            both for ``'post-log'`` and ``'pre-log'`` model.
        """
        global DATA_PATH
        NUM_ANGLES = 1000
        NUM_DET_PIXELS = 513
        self.shape = ((NUM_ANGLES, NUM_DET_PIXELS), (362, 362))
        self.num_elements_per_sample = 2
        if min_pt is None:
            min_pt = MIN_PT
        if max_pt is None:
            max_pt = MAX_PT
        domain = uniform_discr(min_pt, max_pt, self.shape[1], dtype=np.float32)
        if observation_model == 'post-log':
            self.post_log = True
        elif observation_model == 'pre-log':
            self.post_log = False
        else:
            raise ValueError("`observation_model` must be 'post-log' or "
                             "'pre-log', not '{}'".format(observation_model))
        if min_photon_count is None or min_photon_count <= 1.:
            self.min_photon_count = min_photon_count
        else:
            self.min_photon_count = 1.
            warn('`min_photon_count` changed from {} to 1.'.format(
                min_photon_count))
        self.train_len = 35820
        self.validation_len = 3522
        self.test_len = 3553
        self.random_access = True

        while not self.check_for_lodopab():
            print('The LoDoPaB-CT dataset could not be found under the '
                  "configured path '{}'.".format(
                      CONFIG['lodopab_dataset']['data_path']))
            print('Do you want to download it now? (y: download, n: input '
                  'other path)')
            download = input_yes_no()
            if download:
                success = download_lodopab()
                if not success:
                    raise RuntimeError('lodopab dataset not available, '
                                       'download failed')
            else:
                print('Path to LoDoPaB dataset:')
                DATA_PATH = input()
                set_config('lodopab_dataset/data_path', DATA_PATH)

        self.geometry = odl.tomo.parallel_beam_geometry(
            domain, num_angles=NUM_ANGLES, det_shape=(NUM_DET_PIXELS,))
        range_ = uniform_discr(self.geometry.partition.min_pt,
                               self.geometry.partition.max_pt,
                               self.shape[0], dtype=np.float32)
        super().__init__(space=(range_, domain))

    def check_for_lodopab(self):
        """Fast check whether first and last file of each dataset part exist
        under the configured data path.

        Returns
        -------
        exists : bool
            Whether LoDoPaB seems to exist.
        """
        for part in ['train', 'validation', 'test']:
            first_file = os.path.join(
                DATA_PATH, 'observation_{}_000.hdf5'.format(part))
            last_file = os.path.join(
                DATA_PATH, 'observation_{}_{:03d}.hdf5'.format(
                    part, ceil(self.get_len(part) / NUM_SAMPLES_PER_FILE) - 1))
            if not (os.path.exists(first_file) and os.path.exists(last_file)):
                return False
        return True

    def __get_observation_trafo(self, num_samples=1):
        if (self.min_photon_count is None or
                self.min_photon_count == ORIG_MIN_PHOTON_COUNT):
            if self.post_log:
                def observation_trafo(out):
                    pass
            else:
                def observation_trafo(obs):
                    obs *= MU_MAX
                    np.exp(-obs, out=obs)
        else:
            shape = (self.shape[0] if num_samples == 1 else
                     (num_samples,) + self.shape[0])
            mask = np.empty(shape, dtype=np.bool)
            thres0 = 0.5 * (
                -np.log(ORIG_MIN_PHOTON_COUNT/PHOTONS_PER_PIXEL)
                - np.log(1/PHOTONS_PER_PIXEL)) / MU_MAX
            if self.post_log:
                def observation_trafo(obs):
                    np.greater_equal(obs, thres0, out=mask)
                    obs[mask] = -np.log(self.min_photon_count
                                        / PHOTONS_PER_PIXEL) / MU_MAX
            else:
                def observation_trafo(obs):
                    np.greater_equal(obs, thres0, out=mask)
                    obs *= MU_MAX
                    np.exp(-obs, out=obs)
                    obs[mask] = self.min_photon_count/PHOTONS_PER_PIXEL
        return observation_trafo

    def generator(self, part='train'):
        """Yield pairs of low dose observations and (virtual) ground truth.

        Parameters
        ----------
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.

        Yields
        ------
        (observation, ground_truth)
            `observation` : odl element with shape ``(1000, 513)``
                The values depend on the
                `observation_model` and `min_photon_count` parameters that were
                passed to :meth:`__init__`.
            `ground_truth` : odl element with shape ``(362, 362)``
                The values lie in the range ``[0., 1.]``.
        """
        num_files = ceil(self.get_len(part) / NUM_SAMPLES_PER_FILE)
        observation_trafo = self.__get_observation_trafo()
        for i in range(num_files):
            with h5py.File(
                    os.path.join(DATA_PATH, 'ground_truth_{}_{:03d}.hdf5'
                                            .format(part, i)), 'r') as file:
                ground_truth_data = file['data'][:]
            with h5py.File(
                    os.path.join(DATA_PATH, 'observation_{}_{:03d}.hdf5'
                                            .format(part, i)), 'r') as file:
                observation_data = file['data'][:]
            for gt_arr, obs_arr in zip(ground_truth_data, observation_data):
                ground_truth = self.space[1].element(gt_arr)
                observation = self.space[0].element(obs_arr)
                observation_trafo(observation)

                yield (observation, ground_truth)

    def get_ray_trafo(self, impl=None):
        """
        Return the ray transform that is a noiseless version of the forward
        operator.

        Parameters
        ----------
        impl : {``'skimage'``, ``'astra_cpu'``, ``'astra_cuda'``}, optional
            The backend implementation passed to
            :class:`odl.tomo.RayTransform`.

        Returns
        -------
        ray_trafo : odl operator
            The ray transform that corresponds to the noiseless map from
            362 x 362 images to the ``-log`` of their projections (sinograms).
        """
        return odl.tomo.RayTransform(self.space[1], self.geometry,
                                     range=self.space[0], impl=impl)

    ray_trafo = property(get_ray_trafo)
    """
    Read-only property calling :meth:`get_ray_trafo` with default parameters.
    """

    def get_sample(self, index, part='train', out=None):
        """
        Get single sample of the dataset.
        Returns a pair of (virtual) ground truth and its low dose observation,
        of which either part can be left out by option.

        Parameters
        ----------
        index : int
            The index into the dataset part.
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        out : tuple of array-likes or bools, optional
            ``out==(out_observation, out_ground_truth)``

            out_observation : array-like or bool
                Shape ``(1000, 513)``.
                If an odl element or array is passed, the observation is
                written to it.
                If ``True``, a new odl element holding the observation is
                created (the default).
                If ``False``, no observation is returned.
            out_ground_truth : array-like or bool
                Shape ``(362, 362)``.
                If an odl element or array is passed, the ground truth is
                written to it.
                If ``True``, a new odl element holding the ground truth is
                created (the default).
                If ``False``, no ground truth is returned.

        Returns
        -------
        ``(observation, ground_truth)`` or `None`

            observation : odl element or :class:`np.ndarray` or `None`
                Depending on the value of ``out_observation`` (see parameter
                `out`), a newly created odl element, ``out_observation`` or
                `None` is returned.
                The observation values depend on the `observation_model` and
                `min_photon_count` parameters that were given to the
                constructor.

            ground_truth : odl element or :class:`np.ndarray` or `None`
                Depending on the value of ``out_ground_truth`` (see parameter
                `out`), a newly created odl element, ``out_ground_truth`` or
                `None` is returned.
                The values lie in the range ``[0., 1.]``.
        """
        len_part = self.get_len(part)
        if index >= len_part or index < -len_part:
            raise IndexError("index {} out of bounds for part '{}' ({:d})"
                             .format(index, part, len_part))
        if index < 0:
            index += len_part
        (out_observation, out_ground_truth) = out
        file_index = index // NUM_SAMPLES_PER_FILE
        index_in_file = index % NUM_SAMPLES_PER_FILE
        if isinstance(out_observation, bool):
            obs = self.space[0].zero() if out_observation else None
        else:
            obs = out_observation
        if isinstance(out_ground_truth, bool):
            gt = self.space[1].zero() if out_ground_truth else None
        else:
            gt = out_ground_truth
        if obs is not None:
            with h5py.File(
                    os.path.join(DATA_PATH,
                                 'observation_{}_{:03d}.hdf5'
                                 .format(part, file_index)), 'r') as file:
                file['data'].read_direct(np.asarray(obs)[np.newaxis],
                                         np.s_[index_in_file:index_in_file+1],
                                         np.s_[0:1])
            observation_trafo = self.__get_observation_trafo()
            observation_trafo(obs)
        if gt is not None:
            with h5py.File(
                    os.path.join(DATA_PATH,
                                 'ground_truth_{}_{:03d}.hdf5'
                                 .format(part, file_index)), 'r') as file:
                file['data'].read_direct(np.asarray(gt)[np.newaxis],
                                         np.s_[index_in_file:index_in_file+1],
                                         np.s_[0:1])
        return (obs, gt)

    def get_samples(self, key, part='train', out=None):
        """
        Get slice of the dataset.
        Returns a pair of (virtual) ground truth data and its low dose
        observation data, of which either part can be left out by option.

        Parameters
        ----------
        key : slice or range
            The indices into the dataset part.
        part : {``'train'``, ``'validation'``, ``'test'``}, optional
            The data part. Default is ``'train'``.
        out : tuple of arrays or bools, optional
            ``out==(out_observation, out_ground_truth)``

            out_observation : :class:`np.ndarray` or bool
                If an array is passed, the observation data is written to it.
                If ``True``, a new array holding the observation data is
                created (the default).
                If ``False``, no observation data is returned.
            out_ground_truth : :class:`np.ndarray` or bool
                If an array is passed, the ground truth data is written to it.
                If ``True``, a new array holding the ground truth data is
                created (the default).
                If ``False``, no ground truth data is returned.

        Returns
        -------
        ``(observation, ground_truth)`` or `None`

            observation : :class:`np.ndarray` or `None`
                Shape ``(samples, 1000, 513)``.
                Depending on the value of ``out_observation`` (see parameter
                `out`), a newly created array, ``out_observation`` or `None`
                is returned.
                The observation values depend on the `observation_model` and
                `min_photon_count` parameters that were given to the
                constructor.

            ground_truth : :class:`np.ndarray` or `None`
                Shape ``(samples, 362, 362)``.
                Depending on the value of ``out_ground_truth`` (see parameter
                `out`), a newly created array, ``out_ground_truth`` or `None`
                is returned.
                The values lie in the range ``[0., 1.]``.
        """
        len_part = self.get_len(part)
        if isinstance(key, slice):
            key_start = (0 if key.start is None else
                         (key.start if key.start >= 0 else
                          max(0, len_part+key.start)))
            key_stop = (len_part if key.stop is None else
                        (key.stop if key.stop >= 0 else
                         max(0, len_part+key.stop)))
            range_ = range(key_start, key_stop, key.step or 1)
        elif isinstance(key, range):
            range_ = key
        else:
            raise TypeError('`key` expected to have type `slice` or `range`')
        if range_.step < 0:
            raise ValueError('key {} invalid, negative steps are not '
                             'implemented yet'.format(key))
        if range_[-1] >= len_part:
            raise IndexError("key {} out of bounds for part '{}' ({:d})"
                             .format(key, part, len_part))
        range_files = range(range_[0] // NUM_SAMPLES_PER_FILE,
                            range_[-1] // NUM_SAMPLES_PER_FILE + 1)
        (out_observation, out_ground_truth) = out
        # compute slice objects
        slices_files = []
        slices_data = []
        data_count = 0
        for i in range_files:
            if i == range_files.start:
                start = range_.start % NUM_SAMPLES_PER_FILE
            else:
                start = (range_.start - i*NUM_SAMPLES_PER_FILE) % range_.step
            if i == range_files[-1]:
                stop = range_[-1] % NUM_SAMPLES_PER_FILE + 1
            else:
                __next_start = ((range_.start - (i+1)*NUM_SAMPLES_PER_FILE)
                                % range_.step)
                stop = (__next_start - range_.step) % NUM_SAMPLES_PER_FILE + 1
            s = slice(start, stop, range_.step)
            slices_files.append(s)
            len_slice = ceil((s.stop-s.start) / s.step)
            slices_data.append(slice(data_count, data_count+len_slice))
            data_count += len_slice
        # read data
        if isinstance(out_observation, bool):
            obs_arr = (np.empty((len(range_),) + self.shape[0],
                                dtype=np.float32) if out_observation else None)
        else:
            obs_arr = out_observation
        if isinstance(out_ground_truth, bool):
            gt_arr = (np.empty((len(range_),) + self.shape[1],
                               dtype=np.float32) if out_ground_truth else None)
        else:
            gt_arr = out_ground_truth
        if obs_arr is not None:
            for i, slc_f, slc_d in zip(range_files, slices_files, slices_data):
                with h5py.File(
                        os.path.join(DATA_PATH,
                                     'observation_{}_{:03d}.hdf5'
                                     .format(part, i)), 'r') as file:
                    file['data'].read_direct(obs_arr, slc_f, slc_d)
            observation_trafo = self.__get_observation_trafo(
                num_samples=len(obs_arr))
            observation_trafo(obs_arr)
        if gt_arr is not None:
            for i, slc_f, slc_d in zip(range_files, slices_files, slices_data):
                with h5py.File(
                        os.path.join(DATA_PATH,
                                     'ground_truth_{}_{:03d}.hdf5'
                                     .format(part, i)), 'r') as file:
                    file['data'].read_direct(gt_arr, slc_f, slc_d)
        return (obs_arr, gt_arr)

    def create_torch_dataset(self, part='train'):
        import torch

        class LoDoPaBDatasetTorch(Dataset):
            def __init__(self, part='train'):
                self.part = part
                self.dataset = LoDoPaBDataset()

            def __len__(self):
                return self.dataset.get_len(self.part)

            def __getitem__(self, idx):
                obs = np.empty((1,) + self.dataset.shape[0], dtype=np.float32)
                gt = np.empty((1,) + self.dataset.shape[1], dtype=np.float32)
                self.dataset.get_sample(idx, self.part, out=(obs, gt))
                return torch.from_numpy(obs), torch.from_numpy(gt)

        torch_dataset = LoDoPaBDatasetTorch(part=part)
        return torch_dataset
